#!/usr/bin/env python3
# This file is covered by the LICENSE file in the root of this project.
import numpy as np


class LaserScan:
  """Class that contains LaserScan with x,y,z,r"""
  EXTENSIONS_SCAN = ['.bin']

  def __init__(self, project=False, H=64, W=1024, fov_up=3.0, fov_down=-25.0):
    self.project = project
    self.proj_H = H
    self.proj_W = W
    self.proj_fov_up = fov_up
    self.proj_fov_down = fov_down
    self.reset()

  def reset(self):
    """ Reset scan members. """
    self.points = np.zeros((0, 3), dtype=np.float32)        # [m, 3]: x, y, z
    self.remissions = np.zeros((0, 1), dtype=np.float32)    # [m ,1]: remission

    # projected range image - [H,W] range (-1 is no data)
    self.proj_range = np.full((self.proj_H, self.proj_W), -1,
                              dtype=np.float32)

    # unprojected range (list of depths for each point)
    self.unproj_range = np.zeros((0, 1), dtype=np.float32)

    # projected point cloud xyz - [H,W,3] xyz coord (-1 is no data)
    self.proj_xyz = np.full((self.proj_H, self.proj_W, 3), -1,
                            dtype=np.float32)

    # projected remission - [H,W] intensity (-1 is no data)
    self.proj_remission = np.full((self.proj_H, self.proj_W), -1,
                                  dtype=np.float32)

    # projected index (for each pixel, what I am in the pointcloud)
    # [H,W] index (-1 is no data)
    self.proj_idx = np.full((self.proj_H, self.proj_W), -1,
                            dtype=np.int32)

    # for each point, where it is in the range image
    self.proj_x = np.zeros((0, 1), dtype=np.int32)        # [m, 1]: x
    self.proj_y = np.zeros((0, 1), dtype=np.int32)        # [m, 1]: y

    # mask containing for each pixel, if it contains a point or not
    self.proj_mask = np.zeros((self.proj_H, self.proj_W),
                              dtype=np.int32)       # [H,W] mask

  def size(self):
    """ Return the size of the point cloud. """
    return self.points.shape[0]

  def __len__(self):
    return self.size()

  def open_scan(self, filename):
    """ Open raw scan and fill in attributes
    """
    # reset just in case there was an open structure
    self.reset()

    # check filename is string
    if not isinstance(filename, str):
      raise TypeError("Filename should be string type, "
                      "but was {type}".format(type=str(type(filename))))

    # check extension is a laserscan
    if not any(filename.endswith(ext) for ext in self.EXTENSIONS_SCAN):
      raise RuntimeError("Filename extension is not valid scan file.")

    # if all goes well, open pointcloud
    scan = np.fromfile(filename, dtype=np.float32)
    scan = scan.reshape((-1, 4))

    # put in attribute
    points = scan[:, 0:3]    # get xyz
    remissions = scan[:, 3]  # get remission
    self.set_points(points, remissions)

  def set_points(self, points, remissions=None):
    """ Set scan attributes (instead of opening from file)
    """
    # reset just in case there was an open structure
    self.reset()

    # check scan makes sense
    if not isinstance(points, np.ndarray):
      raise TypeError("Scan should be numpy array")

    # check remission makes sense
    if remissions is not None and not isinstance(remissions, np.ndarray):
      raise TypeError("Remissions should be numpy array")

    # put in attribute
    self.points = points    # get xyz
    if remissions is not None:
      self.remissions = remissions  # get remission
    else:
      self.remissions = np.zeros((points.shape[0]), dtype=np.float32)

    # if projection is wanted, then do it and fill in the structure
    if self.project:
      self.do_range_projection()

  def do_range_projection(self):
    """ Project a pointcloud into a spherical projection image.projection.
        Function takes no arguments because it can be also called externally
        if the value of the constructor was not set (in case you change your
        mind about wanting the projection)
    """
    # laser parameters
    fov_up = self.proj_fov_up / 180.0 * np.pi      # field of view up in rad
    fov_down = self.proj_fov_down / 180.0 * np.pi  # field of view down in rad
    fov = abs(fov_down) + abs(fov_up)  # get field of view total in rad

    # get depth of all points
    depth = np.linalg.norm(self.points, 2, axis=1)

    # get scan components
    scan_x = self.points[:, 0]
    scan_y = self.points[:, 1]
    scan_z = self.points[:, 2]

    # get angles of all points
    yaw = -np.arctan2(scan_y, scan_x)
    pitch = np.arcsin(scan_z / depth)

    # get projections in image coords
    proj_x = 0.5 * (yaw / np.pi + 1.0)          # in [0.0, 1.0]
    proj_y = 1.0 - (pitch + abs(fov_down)) / fov        # in [0.0, 1.0]

    # scale to image size using angular resolution
    proj_x *= self.proj_W                              # in [0.0, W]
    proj_y *= self.proj_H                              # in [0.0, H]

    # round and clamp for use as index
    proj_x = np.floor(proj_x)
    proj_x = np.minimum(self.proj_W - 1, proj_x)
    proj_x = np.maximum(0, proj_x).astype(np.int32)   # in [0,W-1]
    self.proj_x = np.copy(proj_x)  # store a copy in orig order

    proj_y = np.floor(proj_y)
    proj_y = np.minimum(self.proj_H - 1, proj_y)
    proj_y = np.maximum(0, proj_y).astype(np.int32)   # in [0,H-1]
    self.proj_y = np.copy(proj_y)  # stope a copy in original order

    # copy of depth in original order
    self.unproj_range = np.copy(depth)

    # order in decreasing depth
    indices = np.arange(depth.shape[0])
    order = np.argsort(depth)[::-1]
    depth = depth[order]
    indices = indices[order]
    points = self.points[order]
    remission = self.remissions[order]
    proj_y = proj_y[order]
    proj_x = proj_x[order]

    # assing to images
    self.proj_range[proj_y, proj_x] = depth
    self.proj_xyz[proj_y, proj_x] = points
    self.proj_remission[proj_y, proj_x] = remission
    self.proj_idx[proj_y, proj_x] = indices
    self.proj_mask = (self.proj_idx > 0).astype(np.int32)


class SemLaserScan(LaserScan):
  """Class that contains LaserScan with x,y,z,r,sem_label,sem_color_label,inst_label,inst_color_label"""
  EXTENSIONS_LABEL = ['.label']

  def __init__(self,  sem_color_dict=None, project=False, H=64, W=1024, fov_up=3.0, fov_down=-25.0, max_classes=300):
    super(SemLaserScan, self).__init__(project, H, W, fov_up, fov_down)
    self.reset()

    # make semantic colors
    if sem_color_dict:
      # if I have a dict, make it
      max_sem_key = 0
      for key, data in sem_color_dict.items():
        if key + 1 > max_sem_key:
          max_sem_key = key + 1
      self.sem_color_lut = np.zeros((max_sem_key + 100, 3), dtype=np.float32)
      for key, value in sem_color_dict.items():
        self.sem_color_lut[key] = np.array(value, np.float32) / 255.0
    else:
      # otherwise make random
      max_sem_key = max_classes
      self.sem_color_lut = np.random.uniform(low=0.0,
                                             high=1.0,
                                             size=(max_sem_key, 3))
      # force zero to a gray-ish color
      self.sem_color_lut[0] = np.full((3), 0.1)

    # make instance colors
    max_inst_id = 100000
    self.inst_color_lut = np.random.uniform(low=0.0,
                                            high=1.0,
                                            size=(max_inst_id, 3))
    # force zero to a gray-ish color
    self.inst_color_lut[0] = np.full((3), 0.1)

  def reset(self):
    """ Reset scan members. """
    super(SemLaserScan, self).reset()

    # semantic labels
    self.sem_label = np.zeros((0, 1), dtype=np.int32)          # [m, 1]: label
    self.sem_label_color = np.zeros((0, 3), dtype=np.float32)  # [m ,3]: color

    # instance labels
    self.inst_label = np.zeros((0, 1), dtype=np.int32)          # [m, 1]: label
    self.inst_label_color = np.zeros((0, 3), dtype=np.float32)  # [m ,3]: color

    # projection color with semantic labels
    self.proj_sem_label = np.zeros((self.proj_H, self.proj_W),
                                   dtype=np.int32)              # [H,W]  label
    self.proj_sem_color = np.zeros((self.proj_H, self.proj_W, 3),
                                   dtype=np.float)              # [H,W,3] color

    # projection color with instance labels
    self.proj_inst_label = np.zeros((self.proj_H, self.proj_W),
                                    dtype=np.int32)              # [H,W]  label
    self.proj_inst_color = np.zeros((self.proj_H, self.proj_W, 3),
                                    dtype=np.float)              # [H,W,3] color

  def open_label(self, filename):
    """ Open raw scan and fill in attributes
    """
    # check filename is string
    if not isinstance(filename, str):
      raise TypeError("Filename should be string type, "
                      "but was {type}".format(type=str(type(filename))))

    # check extension is a laserscan
    if not any(filename.endswith(ext) for ext in self.EXTENSIONS_LABEL):
      raise RuntimeError("Filename extension is not valid label file.")

    # if all goes well, open label
    label = np.fromfile(filename, dtype=np.int32)
    label = label.reshape((-1))

    # set it
    self.set_label(label)

  def set_label(self, label):
    """ Set points for label not from file but from np
    """
    # check label makes sense
    if not isinstance(label, np.ndarray):
      raise TypeError("Label should be numpy array")

    # only fill in attribute if the right size
    if label.shape[0] == self.points.shape[0]:
      self.sem_label = label & 0xFFFF  # semantic label in lower half
      self.inst_label = label >> 16    # instance id in upper half
    else:
      print("Points shape: ", self.points.shape)
      print("Label shape: ", label.shape)
      raise ValueError("Scan and Label don't contain same number of points")

    # sanity check
    assert((self.sem_label + (self.inst_label << 16) == label).all())

    if self.project:
      self.do_label_projection()

  def colorize(self):
    """ Colorize pointcloud with the color of each semantic label
    """
    self.sem_label_color = self.sem_color_lut[self.sem_label]
    self.sem_label_color = self.sem_label_color.reshape((-1, 3))

    self.inst_label_color = self.inst_color_lut[self.inst_label]
    self.inst_label_color = self.inst_label_color.reshape((-1, 3))

  def do_label_projection(self):
    # only map colors to labels that exist
    mask = self.proj_idx >= 0

    # semantics
    self.proj_sem_label[mask] = self.sem_label[self.proj_idx[mask]]
    self.proj_sem_color[mask] = self.sem_color_lut[self.sem_label[self.proj_idx[mask]]]

    # instances
    self.proj_inst_label[mask] = self.inst_label[self.proj_idx[mask]]
    self.proj_inst_color[mask] = self.inst_color_lut[self.inst_label[self.proj_idx[mask]]]



# -*- coding: utf-8 -*-
"""
Created on Fri May  1 14:08:32 2020

@author: arun
"""
import numpy as np
import pandas as pd
import os

class SegCM(object):
    r"""
    Confusion Matix skeleton that carries out basic confusion matrix operations 
    to obtain final confusion matrix
    
    Args:
        nClasses: number of classes 
        labels: target labels
        task: specifies the target task (Object Detection/ Classification) 
    
    """

    def __init__(self, nClasses, labels=None, task="ObjectDetection"):
        self.nClasses = nClasses
        if task == "OD":
            self.CMClasses = (
                self.nClasses + 1
            )  #### Add 1 to represent nothingness as an object
        else:
            self.CMClasses = self.nClasses
        self.CM = np.zeros((self.CMClasses, self.CMClasses), dtype=np.float)
        
        if not labels:
            self.labels = list(range(self.CMClasses))
        else:
            self.labels = labels

    def updateCM(self, groundtruth, predicted):
        """
        Function updates CM based on ground truth and predictions
        Arguments:
            Update vals - array locations of ground truth and predictions on confusion matrix
    
        """
        self.CM[groundtruth][predicted] += 1
        

    def updateincorrectpredictions(self, currCM):
        r"""
        Function adds on to the existing confusion matrix

        """

        self.CM += currCM

    def save_CM(self, outdir, modelname):
        r"""
        Fucntion saves Confusion matrix in target folder
        
        """
        df_cm = pd.DataFrame(self.CM, index=self.labels, columns=self.labels)
        df_cm.to_csv(
            os.path.join(outdir, str(modelname) + "confusion_matrix" + ".csv"),
            index=True,
        )



import os
import json

# ----------------local imports----------------#
from .api_client import APIClient
from .gcp_storage import GCP_Storage
from .console import MainConsole


class AlectioDataset:
    def __init__(self, token: str, root: str, framework: str):
        self.token = token
        self.root = root
        self.framework = framework
        self.console = MainConsole()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(dir_path, "config.json"), "r") as f:
            self.config = json.load(f)
        self.api_client = APIClient(
            backend_url=self.config["backend_ip"], token=self.token)
        
        self.gcp_client = GCP_Storage(exp_token=token)

    def __is_exist(self, dataset_type: str):
        """
        THIS HIDDEN FUNCTION CHECK WHETHER THE DATASET IS EXIST OR NOT IN THE GIVEN ROOT DIR

        Returns:
            _type_: bool
        """
        return os.path.isdir(os.path.join(self.root, dataset_type))

    def __pull_dataset(self, dataset_type: str):
        self.console.info("PULLING DATASET FROM CLOUD")
        error, data = self.api_client.GET_REQUEST(
            end_point=f"/sdk/v2/token/{self.token}/type/{dataset_type}/get_public_dataset",
        )
        if error:
            self.console.error(
                "SOMETHING WENT WRONG WITH ALECTIO PUBLIC DATASET LIBRARY.\nCONTACT ALECTIO ADMIN"
            )

        if data["data"]["task_type"] == "CV":
            os.makedirs(os.path.join(self.root, dataset_type), exist_ok=True)
            self.gcp_client.download_file(
                storage_bucket=data["data"]["storage_bucket"],
                object_key=data["data"]["path"],
                format="zip",
                dest_path=os.path.join(self.root),
            )

        else:
            raise Exception("NLP DATASET IS NOT SUPPORTED YET")

    def __create_dataset(self, dataset_type: str, transforms=None, tf_args=None):
        if self.framework == "pytorch":
            from torchvision.datasets import ImageFolder
            dataset = ImageFolder(
                root=os.path.join(self.root, dataset_type), transform=transforms
            )
            return dataset, len(dataset), dataset.class_to_idx
        elif self.framework == "tensorflow":
            from tensorflow.keras.preprocessing.image import ImageDataGenerator
            data_gen = ImageDataGenerator(**self.transforms)
            dataset = data_gen.flow_from_directory(
                directory=os.path.join(self.root, dataset_type), **tf_args
            )
            return dataset, len(dataset), dataset.class_indices
        else:
            raise ValueError(
                f"Invalid framework {self.framework}, Framework should be one of [tensorflow, pytorch]"
            )

    def get_dataset(self, dataset_type: str, transforms=None, tf_args=None):
        if not self.__is_exist(dataset_type=dataset_type):
            self.__pull_dataset(dataset_type=dataset_type)

        dataset, dataset_len, class_to_idx = self.__create_dataset(
            dataset_type=dataset_type, transforms=transforms, tf_args=tf_args
        )
        return dataset, dataset_len, class_to_idx


# Testing
# if __name__ == "__main__":
#     alectio_dataset = AlectioDataset(
#         token="", root="", framework="pytorch"
#     )
#     train_dataset = alectio_dataset.get_dataset(dataset_type="train", transforms=None)
#     test_dataset = alectio_dataset.get_dataset(dataset_type="test")
#     validation_dataset = alectio_dataset.get_dataset(dataset_type="validation")
#     print(validation_dataset)



import os
import json
from pickle import FALSE, NONE
import requests

from typing import Dict


class APIClient:
    r"""
    A wrapper of python reqeusts module
    It is to handle all GET,POST,PUT,DELETE requests
    """

    def __init__(self, backend_url: str, token: str):
        self.backend_url = backend_url
        self.auth_token = self.AUTH_TOKEN_REQ(token=token)
        self.header = {
            "X-Alectio-Flavor": "PRO",
            "Authorization": f"Bearer {self.auth_token}",
            "Content-Type": "application/json",
        }

    def GET_REQUEST(self, end_point: str, payload: dict = {}):
        url = self.backend_url + end_point

        payload = json.dumps(payload)

        response = requests.request("GET", url, headers=self.header, data=payload)

        if response.status_code not in [200, 201, 202, 203, 204, 2005, 206]:
            return True, None
        else:
            return False, response.json()

    def POST_REQUEST(
        self, end_point: str, payload: dict = {}, auth: dict = {}, headers: dict = {}
    ):
        url = self.backend_url + end_point
        payload = json.dumps(payload)
        response = requests.request("POST", url, headers=self.header, data=payload)

        if response.status_code not in [200, 201, 202, 203, 204, 2005, 206]:
            return True, None
        else:
            return False, response.json()

    def AUTH_TOKEN_REQ(self, token):
        url = self.backend_url + f"/api/v2/python_client_auth/experiment_token/{token}"
        payload = {}
        headers = {"X-Alectio-Flavor": "PRO"}
        response = requests.request("GET", url, headers=headers, data=payload).json()
        return response["data"]["access_token"]



# hardcoded task type as image_classification for testing
import numpy as np
import json
import os
import time
import ast
from scipy.special import softmax
import datetime
from .utils import seed_comparison
from typing import Dict
from copy import deepcopy
from collections import Counter
from .api_client import APIClient
from .multiple_seed_pool import DataPool
from alectio_sdk.metrics.object_detection import Metrics, batch_to_numpy
from alectio_sdk.metrics.classification import ClassificationMetrics
from alectio_sdk.metrics.regression import RegressionMetrics
from alectio_sdk.metrics.object_segmentation import segmentation_metrics
from alectio_sdk.metrics.multilabel_classification import (
    MultiLabelClassificationMetrics,
)
from codecarbon import OfflineEmissionsTracker as EmissionsTracker
import pandas as pd
from alectio_sdk.metrics.object_segmentation import SegMetrics
from rich.progress import track
from alectio_sdk.sdk.sql_client import create_connection
import logging
from .console import MainConsole
from .gcp_storage import GCP_Storage
import requests

import warnings

warnings.filterwarnings("ignore")


class Pipeline(object):
    r"""
    A wrapper for your `train`, `test`, and `infer` function. The arguments for your functions should be specifed
    separately and passed to your pipeline object during creation.

    Args:
        name (str): experiment name
        train_fn (function): function to be executed in the train cycle of the experiment.
        test_fn (function): function to be executed in the test cycle of the experiment.
        infer_fn (function): function to be executed in the inference cycle of the experiment.
        getstate_fn (function): function specifying a mapping between indices and file names.


    """

    def __init__(
        self,
        name,
        train_fn,
        test_fn,
        infer_fn,
        getstate_fn,
        args,
        token,
        multiple_initialisations={"seeds": [], "limit_value": 0},
    ):
        self.console = MainConsole()
        # time saved functionality
        self.time_saved_dir = f"{args['LOG_DIR']}/time_saved"
        self.time_saved_file = "time_saved.csv"
        if not os.path.exists(self.time_saved_dir):
            os.mkdir(self.time_saved_dir)

        # tracking co2_emissions
        self.co2_dir = f"{args['LOG_DIR']}/co2"
        if not os.path.exists(self.co2_dir):
            os.mkdir(self.co2_dir)
        self.co2_tracker = EmissionsTracker(
            country_iso_code="CAN", output_dir=self.co2_dir, log_level="critical"
        )

        self.train_fn = train_fn
        self.test_fn = test_fn
        self.infer_fn = infer_fn
        self.getstate_fn = getstate_fn
        self.args = args
        # new arg

        # specifically for multiple seeds
        self.multiple_seeds = multiple_initialisations["seeds"]
        self.limit_value = int(multiple_initialisations["limit_value"])
        self.best_seed = 42

        self.client = GCP_Storage(token)

        dir_path = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.join(dir_path, "config.json"), "r") as f:
            self.config = json.load(f)

        self.client_token = token

        self.labeled = []
        self._testUploaded = False
        self._uploaded_infer_index = []

        self.api_client = APIClient(
            backend_url=self.config["backend_ip"], token=self.client_token
        )
        self.auth_token = self.api_client.AUTH_TOKEN_REQ(self.client_token)
        # self.console.info("ALECTIO PIPELINE RECEIVED ENCRYPTED TOKEN")

    def notify_sdk_status(self, logdir):
        logging.basicConfig(
            filename=os.path.join(logdir, "Appstatus.log"), level=logging.INFO
        )
        self.console.success("SDK Alectio initialized successfully")
        self.console.message(
            "Training checkpoints and other logs for current experiment will be written into the folder {}".format(
                logdir
            )
        )
        self.console.message("Press CTRL + C to exit")

    def _checkdirs(self, dir_):
        if not os.path.exists(dir_):
            os.makedirs(dir_, exist_ok=True)

    def check_model_activations(self, model_outputs):
        supported_activations = {"pre_softmax", "softmax", "sigmoid", "logits"}
        activation_set = set()
        for k, v in model_outputs.items():
            if isinstance(v, dict):
                activation_set.update(set(v.keys()).intersection(supported_activations))
            if len(activation_set) == 0:
                raise TypeError("Incorrect logit output format encountered")

        return activation_set

    def _estimate_exp_time(self, last_time):
        """
        Estimates the compute time remaining for the experiment

        Args:
            train_times (list): training_times noted down so far
            n_loop (int): total number of loops
        """

        def convert(seconds):
            seconds = seconds % (24 * 3600)
            hour = seconds // 3600
            seconds %= 3600
            minutes = seconds // 60
            seconds %= 60

            return "%d:%02d:%02d" % (hour, minutes, seconds)

        loops_completed = self.current_loop + 1
        time_left = convert(last_time * (self.n_loop - loops_completed))
        self.console.message(
            "Estimated time left for the experiment: {}".format(time_left)
        )
        return time_left

    def _estimate_seed_init_time(self, last_time):
        pass

    def time_saved(
        self, epoch, start_train, end_train, total_records, selected_records
    ):
        actual_train_time = end_train - start_train
        estimated_train_time = total_records * (actual_train_time / selected_records)
        total_time_saved = estimated_train_time - actual_train_time
        time_saved_file_path = os.path.join(self.time_saved_dir, self.time_saved_file)

        info = {
            "epoch": [epoch],
            "selected_records": [selected_records],
            "total_records": [total_records],
            "training_time": [actual_train_time],
            "estimated_train_time": [estimated_train_time],
            "time_saved": [total_time_saved],
        }

        if os.path.exists(time_saved_file_path):
            df = pd.read_csv(time_saved_file_path)
            last_epoch = df["epoch"].to_list()[-1]
            if last_epoch == epoch:
                data = [x[0] for x in list(info.values())]
                df.loc[epoch] = data
            else:
                new_df = pd.DataFrame(info)
                df = pd.concat([df, new_df], axis=0)
            df.to_csv(time_saved_file_path, index=False)
        else:
            df = pd.DataFrame(info)
            df.to_csv(time_saved_file_path, index=False)

        return

    def one_loop(self, request):
        # Get payload args

        self.console.message("Extracting payload arguments from Alectio")

        payload = {
            "tenant_id": request["tenant_id"],
            "experiment_id": request["experiment_id"],
            "project_id": request["project_id"],
            "current_loop": request["current_loop"],
            "bucket_name": request["bucket_name"],
            "task_type": request["task_type"].lower(),
            "dataset_type": request["dataset_type"].lower(),
            "training_set_size": request["training_set_size"],
            "n_rec_per_loop": request["n_rec_per_loop"],
            "n_loop": request["n_loop"],
            "class_labels": request["classes"],
            "num_of_classes": request["num_of_classes"],
        }
        self.logdir = payload["experiment_id"]
        self._checkdirs(self.logdir)
        self.training_set_size = payload["training_set_size"]
        self.notify_sdk_status(self.logdir)

        self.console.success("Valid payload arguments extracted")
        self.console.message("Initializing process to train and optimize your model")
        returned_payload = self.run_loop(payload, self.args)

        self.console.success("Optimization process complete !")
        self.console.info(
            "Results of the loop will be available in Alectio Platform soon"
        )

        error, response = self.api_client.POST_REQUEST(
            end_point="/sdk/v2/end_of_task",
            payload={
                "experiment_id": request["experiment_id"],
                "project_id": request["project_id"],
                "num_of_classes": request["num_of_classes"],
            },
        )
        if error:
            self.console.warning(
                "SOMETHING WENT WRONG WITH YOUR INFER FUNCTION !!\nRECHECK YOUR CODE OR REACH OUT TO ALECTIO ADMIN FOR ANY SUPPORT"
            )
            return {"Message": "Loop Failed - non 200 status returned"}

        else:
            self.console.message(
                "Experiment {} running".format(payload["experiment_id"])
            )
            return {"Message": "Loop Started - 200 status returned"}

    def run_loop(self, payload, args):
        r"""
        Executes one loop of active learning. Returns the read `payload` back to the user.

        Args:
           args: a dict with the key `sample_payload` (required path) and any arguments needed by the `train`, `test`
           and infer functions.
        Example::

            args = {sample_payload: 'sample_payload.json', EXPT_DIR : "./log", exp_name: "test", \
                                                                 train_epochs: 1, batch_size: 8}


        """
        self.experiment_id = payload["experiment_id"]
        self.project_id = payload["project_id"]
        self.tenant_id = payload["tenant_id"]
        self.console.info("Extracting essential experiment params")

        # read selected indices upto this loop
        payload["current_loop"] = int(payload["current_loop"])
        self.current_loop = payload["current_loop"]
        self.bucket_name = payload["bucket_name"]
        self.n_loop = payload["n_loop"]
        self.dataset_type = payload["dataset_type"].lower()
        self.class_labels = payload["class_labels"]
        self.num_of_classes = payload["num_of_classes"]
        self.training_set_size = payload["training_set_size"]

        # type of the ML problem

        self.task_type = payload["task_type"]
        self.task_type = self.task_type.lower()

        self.expt_dir = os.path.join(
            self.tenant_id, self.project_id, self.experiment_id
        )
        self.expt_dir = f'{self.tenant_id}/{self.project_id}/{self.experiment_id}'

        self.console.success("Essential experiment params have been extracted")

        self.console.info("Verifying the meta.json that was uploaded by the user")

        self.labeled = []
        self.console.info("Reading indices to train on")

        # for i in range(self.current_loop + 1):
        for i in range(self.current_loop + 1):  #! LOOK AT HERE ONCE YOU WILL COME BACK
            # object_key = os.path.join(
            #     self.expt_dir, "selected_indices_{}.pkl".format(i))
            try:
                object_key = f'{self.expt_dir}/selected_indices_{i}.pkl'
                selected_indices = self.client.read_file(
                    storage_bucket=self.bucket_name, object_key=object_key, format="pkl"
                
                
                )
            except Exception as e:
                print(e)                
            # print(self.expt_dir)
            # print(selected_indices)
            self.labeled.extend(selected_indices)
            # print("inside one loop function")
            # print(self.labeled)

        if self.current_loop == 0:
            if self.multiple_seeds != []:
                initial_unlabeled = [i for i in range(self.limit_value)]
                data_pool = DataPool(unlabeled=initial_unlabeled)
                self.resume_from = None
                best_seed_info = dict()
                best_seed_info["seed"] = self.multiple_seeds[0]
                best_seed_info["metric"] = 0

                for seed in self.multiple_seeds:
                    # data generation code
                    self.labeled = list()
                    n_rec = self.limit_value // 10
                    selected_indices = data_pool.random_sample("unlabeled", n_rec, seed)
                    self.labeled.extend(selected_indices)

                    # train and test
                    self.train_seed(args, seed)
                    self.console.message("Initializing testing of your model !")
                    metrics = self.test_seed(args, seed)
                    self.console.success(f"Testing finished for seed {seed}!")

                    # compare metrics
                    new_seed_info = dict()
                    new_seed_info["seed"] = seed
                    if self.task_type == "object_detection":
                        new_seed_info["metric"] = metrics["mAP"]
                    elif self.task_type == "regression":
                        new_seed_info["metric"] = metrics["r2_score"]
                    else:
                        new_seed_info["metric"] = metrics["accuracy"]

                    best_seed_info = seed_comparison(best_seed_info, new_seed_info)

                self.best_seed = best_seed_info["seed"]

            # send seed to sdk-backend
            url = "https://api-v2-pro.alectio.com/sdk/v2/seed_value"

            payload = json.dumps({"seed_value": self.best_seed})
            headers = {
                "X-Alectio-Flavor": "PRO",
                "Authorization": "Bearer " + self.auth_token,
                "Content-Type": "application/json",
            }
            response = requests.request("POST", url, headers=headers, data=payload)

            # Normal Flow
            self.resume_from = None
            self.console.info(
                "Extracting indices for our reference, this may take time ... Please be patient"
            )
            self.state_json = self.getstate_fn(args)
            object_key = os.path.join(self.expt_dir, "data_map.pkl")
            object_key = f"{self.expt_dir}/data_map.pkl"
            self.console.success("Extraction complete !!!")
            self.console.message(
                "Creating index to records data reference for the current experiment"
            )
            self.client.upload_object(
                object=self.state_json,
                storage_bucket=self.bucket_name,
                object_key=object_key,
                format="pkl",
            )
            self.console.success("Reference creation complete")

        else:
            self.console.info("Resuming from a checkpoint from a previous loop ")
            # two dag approach needs to refer to the previous checkpoint
            self.resume_from = "ckpt_{}".format(self.current_loop - 1)

        self.ckpt_file = "ckpt_{}".format(self.current_loop)
        self.console.message("Initializing training of your model")

        # time saved functionality
        train_start_time = time.time()
        self.train(args)
        train_end_time = time.time()
        self.time_saved(
            # self,
            epoch=self.current_loop,
            start_train=train_start_time,
            end_train=train_end_time,
            total_records=self.training_set_size,
            selected_records=len(self.labeled),
        )
        self.console.success("Training finished !")
        self.console.message("Initializing testing of your model !")
        self.test(args)
        self.console.success("Testing finished !")
        self.console.message("Assessing current best model")
        self.infer(args)
        self.console.success("Assessment complete ")
        self.console.info(
            "Time to check what records to train on next loop , visit our front end for more details"
        )

        return payload

    def train(self, args):
        r"""
        A wrapper for your `train` function. Returns `None`.

        Args:
        args: a dict whose keys include all of the arguments needed for your `train` function which is defined in `processes.py`.

        """

        def get_hyperparams(hyperparameters: Dict):
            extracted_hyp = dict()

            hyp_parameter_names = [
                "optimizer_name",
                "loss",
                "running_loss",
                "epochs",
                "batch_size",
                "loss_function",
                "activation",
                "optimizer",
            ]

            for param in hyp_parameter_names[:-1]:
                if param in hyperparameters.keys():
                    extracted_hyp[param] = hyperparameters[param]
                else:
                    extracted_hyp[param] = None

            # optimizer
            if "optimizer" in hyperparameters.keys():
                try:
                    optimizer_params = hyperparameters["optimizer"]["param_groups"][0]
                    extracted_hyp["optimizer"] = optimizer_params
                except:
                    extracted_hyp["optimizer"] = hyperparameters["optimizer"]
            else:
                extracted_hyp["optimizer"] = None

            extracted_hyp["time"] = datetime.datetime.now()
            return extracted_hyp

        start = time.time()

        if self.labeled is not None and len(self.labeled) == 0:
            # print(self.labeled)
            raise ValueError("Labeled indices from backend are empty or None")

        self.console.info("Labeled records are ready to be trained")
        self.labeled.sort()  # Maintain increasing order

        # if len(self.labeled) > len(set(self.labeled)):
        #     raise ValueError("There exist repeated records.")

        all_labeled = dict(Counter(self.labeled))
        duplicates = []
        for k, v in all_labeled.items():
            if v == 2:
                duplicates.append(k)

        self.labeled = list(set(self.labeled))

        train_op = self.train_fn(
            args,
            labeled=deepcopy(self.labeled),
            resume_from=self.resume_from,
            ckpt_file=self.ckpt_file,
        )
        extracted_hyperparams = {}
        if train_op is not None:
            labels = train_op["labels"]
            unique, counts = np.unique(labels, return_counts=True)
            num_queried_per_class = {u: c for u, c in zip(unique, counts)}

            if "hyperparams" in train_op.keys():
                extracted_hyperparams = get_hyperparams(train_op["hyperparams"])

        end = time.time()

        # @TODO compute insights from labels
        if train_op is not None:
            insights = {
                "train_time": end - start,
                "num_queried_per_class": num_queried_per_class,
            }
        else:
            insights = {"train_time": end - start}

        self._estimate_exp_time(insights["train_time"])
        object_key = os.path.join(
            self.expt_dir, "insights_{}.pkl".format(self.current_loop)
        )
        object_key = f"{self.expt_dir}/insights_{self.current_loop}.pkl"
        

        self.client.upload_object(
            object=insights,
            storage_bucket=self.bucket_name,
            object_key=object_key,
            format="pkl",
        )

        object_key = f"{self.expt_dir}/meta_temp_{self.current_loop}.pkl"
        
        self.client.upload_object(
            object=extracted_hyperparams,
            storage_bucket=self.bucket_name,
            object_key=object_key,
            format="pkl",
        )

        return

    def train_seed(self, args, seed):
        r"""
        A wrapper for your `train` function. Used for training on mutiple seeds. Returns `None`.

        Args:
        args: a dict whose keys include all of the arguments needed for your `train` function which is defined in `processes.py`.

        """
        start = time.time()

        if self.labeled is not None and len(self.labeled) == 0:
            raise ValueError("Labeled indices from backend are empty or None")

        self.console.info("Labeled records are ready to be trained")
        self.labeled.sort()  # Maintain increasing order

        all_labeled = dict(Counter(self.labeled))
        duplicates = []
        for k, v in all_labeled.items():
            if v == 2:
                duplicates.append(k)

        self.labeled = list(set(self.labeled))

        train_op = self.train_fn(
            args,
            labeled=deepcopy(self.labeled),
            resume_from=self.resume_from,
            ckpt_file=self.ckpt_file,
        )

        if train_op is not None:
            labels = train_op["labels"]
            unique, counts = np.unique(labels, return_counts=True)
            num_queried_per_class = {u: c for u, c in zip(unique, counts)}

        end = time.time()

        # @TODO compute insights from labels
        if train_op is not None:
            insights = {
                "train_time": end - start,
                "num_queried_per_class": num_queried_per_class,
            }
        else:
            insights = {"train_time": end - start}

        self._estimate_exp_time(insights["train_time"])
        self.console.info(f"Training on seed: {seed} completed!")

        return

    def test(self, args):
        """
        A wrapper for your `test` function which writes predictions and ground truth to the specified S3 bucket. Returns `None`.

        Args:
        args: a dict whose keys include all of the arguments needed for your `test` function which is defined in `processes.py`.

        """
        self.console.info("Extracting test results ")

        # TODO: UPLOAD TEST DATA (IT IS OPTIONAL)
        """
            1. TAKE IMAGE/TEXT AS INPUT AND RETURN FROM TEST FUNCTION
            2. UPLOAD THE LABEL DATA
                I. IF TEXT DATA PUT IT INTO .PKL/ES
                II. IF IMAGE DATA, PUT IT TO S3 
            3. TEST DATA MAPPING
            4. SET IS_TEST UPLOAD TO TRUE(FOR EACH INITIAL CALL OF PIPELINE, IS_UPLOADTEST IS FALSE)
        
        """

        res = self.test_fn(args, ckpt_file=self.ckpt_file)

        # res = self.test_fn(args, ckpt_file=self.ckpt_file)

        if self.task_type == "3d_segmentation":
            if "rangebased" in res:
                (
                    _3D_predictions,
                    _3D_labels,
                    _2D_predictions,
                    _2D_labels,
                    range_filtered,
                    range_default,
                    input_data,
                ) = (
                    res["3Dpredictions"],
                    res["3Dlabels"],
                    res["2Dpredictions"],
                    res["2Dlabels"],
                    res["rangebased"],
                    res["default"],
                    res["input_data"] if "input_data" in res else [],
                )
            else:
                (
                    _3D_predictions,
                    _3D_labels,
                    _2D_predictions,
                    _2D_labels,
                    range_filtered,
                    range_default,
                    input_data,
                ) = (
                    res["3Dpredictions"],
                    res["3Dlabels"],
                    res["2Dpredictions"],
                    res["2Dlabels"],
                    None,
                    None,
                    res["input_data"] if "input_data" in res else [],
                )
        else:
            predictions, ground_truth, input_data = (
                res["predictions"],
                res["labels"],
                res["input_data"] if "input_data" in res else [],
            )

        print("#" * 100)
        print(f"TOTAL IMAGES TO UPLOAD --------------------->{len(input_data)}")
        print("#" * 100)
        self.console.message("Writing test results to BUCKET")

        # UPLOAD LABEL DATA TO S3
        """
        IF TEXT DATA:
            MAKE A DICTIONARY TO SAVE INDEX AND TEXT IN A .PKL FILE
        IF IMAGE DATA:
            MAKE A DICTIONARY TO SAVE INDEX AND IMAGE S3 KEY IN A .PKL FILE
        """

        if self.current_loop == 0 and len(input_data) > 0:
            msg = (
                "ℹ️  "
                + "Test data will be uploaded whenever at Loop 0. \nYou can stop it from your code(Recommended to pass test data for better understanding)"
                + "\n"
            )

            self.console.message(msg)
            if "image" in self.dataset_type:
                data_map = []
                for i, image_array in enumerate(
                    track(input_data, description="Uploading test data...")
                ):
                    image_key = os.path.join(self.expt_dir, "test", str(i) + ".png")
                    data_map.append(
                        {
                            "index": i,
                            "image_path": image_key,
                            # "project_id": self.project_id,
                            # "experiment_id": self.experiment_id,
                            # "datamap_type": "test_datamap",
                        }
                    )

                    self.client.upload_object(
                        storage_bucket=self.bucket_name,
                        object=image_array,
                        object_key=image_key,
                        format="png",
                    )

                # backup
                object_key = os.path.join(
                    self.expt_dir, "test_datamap.json".format(self.current_loop)
                )
                self.client.upload_object(
                    object=data_map,
                    storage_bucket=self.bucket_name,
                    object_key=object_key,
                    format="pkl",
                )
                # self._testUploaded = True

            # if "text" in self.dataset_type:
            #     data_map = []
            #     for i, label in enumerate(input_data):
            #         data_map.append(
            #             {
            #                 "index": i,
            #                 "text": label,
            #                 "project_id": self.project_id,
            #                 "experiment_id": self.experiment_id,
            #                 "datamap_type": "test_datamap",
            #             }
            #         )

            #     # backup data
            #     object_key = os.path.join(
            #         self.expt_dir, "test_datamap.pkl".format(self.current_loop)
            #     )
            #     self.client.upload_object(
            #         predictions, self.bucket_name, object_key, "pickle"
            #     )
            #     # ES storing
            #     # Feature will be released in V2 with lake
            #     # es_upload_datmap(datamap=data_map)
            # self._testUploaded = True

        # write predictions and labels to S3
        # object_key = os.path.join(
        #     self.expt_dir, "test_predictions_{}.pkl".format(self.current_loop)
        # )
        # self.client.upload_object(
        #     object=predictions,
        #     storage_bucket=self.bucket_name,
        #     object_key=object_key,
        #     format="pkl",
        # )

        # if self.current_loop == 0:
        #     # write ground truth to S3
        #     object_key = os.path.join(
        #         self.expt_dir, "test_ground_truth.pkl".format(self.current_loop)
        #     )
        #     self.client.upload_object(
        #         object=ground_truth,
        #         storage_bucket=self.bucket_name,
        #         object_key=object_key,
        #         format="pkl",
        #     )

        if "3D" in self.dataset_type:
            self.compute3D_metrics(
                _3D_predictions, _3D_labels, _2D_predictions, _2D_labels, range_filtered
            )
        else:
            self.compute_metrics(predictions, ground_truth)
        return

    def test_seed(self, args, seed):
        """
        A wrapper for your `test` function which writes predictions and ground truth to the specified S3 bucket. Returns `metrics`.
        Used for training on mutiple seeds.

        Args:
        args: a dict whose keys include all of the arguments needed for your `test` function which is defined in `processes.py`.

        """
        self.console.info(f"Extracting test results for seed: {seed}")

        # TODO: UPLOAD TEST DATA (IT IS OPTIONAL)
        """
            1. TAKE IMAGE/TEXT AS INPUT AND RETURN FROM TEST FUNCTION
            2. UPLOAD THE LABEL DATA
                I. IF TEXT DATA PUT IT INTO .PKL/ES
                II. IF IMAGE DATA, PUT IT TO S3 
            3. TEST DATA MAPPING
            4. SET IS_TEST UPLOAD TO TRUE(FOR EACH INITIAL CALL OF PIPELINE, IS_UPLOADTEST IS FALSE)
        
        """

        res = self.test_fn(args, ckpt_file=self.ckpt_file)

        if self.task_type == "3d_segmentation":
            if "rangebased" in res:
                (
                    _3D_predictions,
                    _3D_labels,
                    _2D_predictions,
                    _2D_labels,
                    range_filtered,
                    range_default,
                    input_data,
                ) = (
                    res["3Dpredictions"],
                    res["3Dlabels"],
                    res["2Dpredictions"],
                    res["2Dlabels"],
                    res["rangebased"],
                    res["default"],
                    res["input_data"] if "input_data" in res else [],
                )
            else:
                (
                    _3D_predictions,
                    _3D_labels,
                    _2D_predictions,
                    _2D_labels,
                    range_filtered,
                    range_default,
                    input_data,
                ) = (
                    res["3Dpredictions"],
                    res["3Dlabels"],
                    res["2Dpredictions"],
                    res["2Dlabels"],
                    None,
                    None,
                    res["input_data"] if "input_data" in res else [],
                )
        else:
            predictions, ground_truth, input_data = (
                res["predictions"],
                res["labels"],
                res["input_data"] if "input_data" in res else [],
            )

        if "3D" in self.dataset_type:
            metrics = self.compute3D_metrics(
                _3D_predictions,
                _3D_labels,
                _2D_predictions,
                _2D_labels,
                range_filtered,
                upload_metrics=False,
                return_metrics=True,
            )
        else:
            metrics = self.compute_metrics(
                predictions, ground_truth, upload_metrics=False, return_metrics=True
            )

        return metrics

    # correct the self.type when confirmed.
    def compute3D_metrics(
        self,
        pclpredictions,
        pcllabels,
        imgpredictions,
        imglabels,
        range_filtered,
        range_default,
        upload_metrics=True,
        return_metrics=False,
    ):
        metrics = {}
        if self.task_type == "3d_object_detection":
            raise NotImplementedError(
                "3D object detection evaluation has not been implemented yet"
            )

        elif self.task_type == "3d_segmentation":
            m = SegMetrics(
                n_classes=self.num_of_classes,
                labels=ast.literal_eval(self.class_labels),
                return_2D=True,
                return_3D=True,
                rangenet=True,
                default_ranges=range_default,
                include_ranges=True,
                range_labels=range_filtered,
            )

            m.evaluate3D(pcllabels, pclpredictions, rangelabels=range_filtered)
            m.evaluate2D(imglabels, imgpredictions)
            metrics = {
                "3DCM": m.get3DCM(),
                "3DrangeCM": m.get3DrangeCM(),
                "3DmIOU": m.get3DmIOU(),
                "3DrangemIOU": m.get3DrangemIOU(),
                "3DIOU": m.get3DIOU(),
                "3Drange3DIOU": m.get3DrangeIOU(),
                "3DDICE": m.get3DDICE(),
                "3DmDICE": m.get3DmDICE(),
                "3DrangeDICE": m.get3DrangeDICE(),
                "3DrangemDICE": m.get3DrangemDICE(),
                "3DAcc": m.get3Dacc(),
                "3DmAcc": m.get3Dmacc(),
                "3DrangeAcc": m.get3Drangeacc(),
                "3DrangemAcc": m.get3Drangemacc(),
                "class_labels": ast.literal_eval(self.class_labels),
            }

        # save metrics to S3
        if upload_metrics:
            object_key = os.path.join(
                self.expt_dir, "metrics_{}.pkl".format(self.current_loop)
            )
            self.client.upload_object(
                object=metrics,
                storage_bucket=self.bucket_name,
                object_key=object_key,
                format="pkl",
            )

        if return_metrics:
            return metrics

        return

    def compute_metrics(
        self, predictions, ground_truth, upload_metrics=True, return_metrics=False
    ):
        metrics = {}
        if self.task_type == "object_detection":
            det_boxes, det_labels, det_scores, true_boxes, true_labels = batch_to_numpy(
                predictions, ground_truth
            )

            m = Metrics(
                det_boxes=det_boxes,
                det_labels=det_labels,
                det_scores=det_scores,
                true_boxes=true_boxes,
                true_labels=true_labels,
                num_classes=len(self.class_labels),
            )

            metrics = {
                "mAP": m.getmAP(),
                "AP": m.getAP(),
                "precision": m.getprecision(),
                "recall": m.getrecall(),
                "confusion_matrix": m.getCM().tolist(),
                "class_labels": self.class_labels,
            }

        elif (
            (
                "single_label_classification" in self.task_type
                and self.dataset_type == "text"
            )
            or (
                "single_label_classification" in self.task_type
                and self.dataset_type == "2d_image"
            )
            or (
                "single_label_classification" in self.task_type
                and self.dataset_type == "numeric"
            )
        ):
            m = ClassificationMetrics(predictions, ground_truth)

            metrics = {
                "accuracy": m.get_accuracy(),
                "f1_score_per_class": m.get_f1_score_per_class(),
                "f1_score": m.get_f1_score(),
                "precision_per_class": m.get_precision_per_class(),
                "precision": m.get_precision(),
                "recall_per_class": m.get_recall_per_class(),
                "recall": m.get_recall(),
                "confusion_matrix": m.get_confusion_matrix(),
                "acc_per_class": m.get_acc_per_class(),
                "label_disagreement": m.get_label_disagreement(),
                "confusion_matrix_per_class": m.get_confusion_matrix_per_class(),
            }

        elif (
            self.task_type == "instance_segmentation"
            and self.dataset_type == "2d_image"
        ):
            metrics = segmentation_metrics(
                ground_truth,
                predictions,
                n_classes=self.num_of_classes,
            )

        elif (
            "multi_label_classification" in self.task_type
            and self.dataset_type == "text"
        ) or (
            "multi_label_classification" in self.task_type
            and self.dataset_type == "2d_image"
        ):
            m = MultiLabelClassificationMetrics(predictions, ground_truth)

            metrics = {
                "accuracy": m.get_accuracy(),
                "micro_f1": m.get_f1_score_micro(),
                "macro_f1": m.get_f1_score_macro(),
                "micro_precision": m.get_precision_micro(),
                "macro_precision": m.get_precision_macro(),
                "micro_recall": m.get_recall_micro(),
                "macro_recall": m.get_recall_macro(),
                "multilabel_confusion_matrix": m.get_confusion_matrix(),
                "hamming_loss": m.get_hamming_loss(),
            }
        elif self.task_type == "regression" and self.dataset_type == "numeric":
            r = RegressionMetrics(target=ground_truth, prediction=predictions)

            metrics = {
                "mean_absolute_error": r.get_MAE(),
                "mean_squared_error": r.get_MSE(),
                "root_mean_squared_error": r.get_RSME(),
                "root_mean_squared_log_error": r.get_RMSLE(),
                "r2_score": r.get_R2(),
                "adjusted_r2_score": r.get_adjusted_R2(),
            }

        else:
            raise ValueError(
                f"The selected task type is currently not supported, received type : {self.task_type}"
            )

        self.console.info(f"CURRENT LOOP---------------------->{self.current_loop}")
        # save metrics to S3
        if upload_metrics:
            #todo: check the below
            object_key = os.path.join(
                self.expt_dir, "metrics_{}.pkl".format(self.current_loop)
            )
            object_key = f'{self.expt_dir}/metrics_{self.current_loop}.pkl'
            self.console.warning(f"STORING METRIC FILE TO {object_key}")
            self.client.upload_object(
                object=metrics,
                storage_bucket=self.bucket_name,
                object_key=object_key,
                format="pkl",
            )

        if return_metrics:
            return metrics

        return

    def pre_softmax_to_softmax(self, pre_softmax_values):
        exp_values = np.exp(pre_softmax_values)
        softmax_values = exp_values / np.sum(exp_values)
        return softmax_values

    def logit_to_softmax(self, logit_values):
        return softmax(logit_values, dim=-1)

    def sigmoid_to_softmax(sigmoid_values):
        sigmoid_values = np.array(sigmoid_values)
        logits = np.log(sigmoid_values / (1 - sigmoid_values))
        exp_logits = np.exp(logits)
        softmax_values = exp_logits / (exp_logits + 1)
        return softmax_values

    def calculate_confidence(self, pre_softmax_values):
        return np.max(pre_softmax_values)

    def calculate_entropy(self, pre_softmax_values):
        exp_values = np.exp(pre_softmax_values)
        prob_distribution = exp_values / np.sum(exp_values)
        entropy = -np.sum(prob_distribution * np.log(prob_distribution))
        return entropy

    def calculate_margin(self, pre_softmax_values):
        max_value = np.max(pre_softmax_values)
        second_max_value = np.partition(pre_softmax_values, -2)[-2]
        margin = max_value - second_max_value
        return margin

    def infer(self, args):
        r"""
        A wrapper for your `infer` function which writes outputs to the specified S3 bucket. Returns `None`.

        Args:
            args: a dict whose keys include all of the arguments needed for your `infer` function which is defined in `processes.py`.

        """
        self.console.message(
            "Getting insights on currently unused/unlabelled train data"
        )
        self.console.message("This may take some time. Please be patient ............")

        ts = range(int(self.training_set_size))
        self.unlabeled = sorted(list(set(ts) - set(self.labeled)))
        args["current_loop"] = self.current_loop
        outputs = self.infer_fn(
            args, unlabeled=deepcopy(self.unlabeled), ckpt_file=self.ckpt_file
        )
        local_db = os.path.join(
            self.args["EXPT_DIR"], "infer_outputs_{}.db".format(self.current_loop)
        )

        if os.path.exists(local_db) or outputs is not None:
            self.console.info(
                "Sending assessments on unlabelled train set to Alectio team"
            )
            self.console.warning(
                f"self.task_type--------------------->{self.task_type}"
            )
            if "classification" in self.task_type:
                outputs = outputs["outputs"]
                # finds correct activation as key i.e. 'sigmoid' or 'logits'
                activation = self.check_model_activations(outputs)

                # Remap to absolute indices
                logits_dict = {}
                remap_dict = {}
                try:
                    object_key=os.path.join(
                            self.expt_dir, "confidence_logit_entropy.pkl"
                        )
                    object_key = f"{self.expt_dir}/confidence_logit_entropy.pkl"
                    classification_distribution = self.client.read_file(
                        storage_bucket=self.bucket_name,
                        object_key=object_key
                        ,
                        format="pkl",
                    )
                except:
                    classification_distribution = {}
                # confidence_vals = []
                for key in outputs.keys():
                    remap_dict[key] = self.unlabeled.pop(0)

                while len(outputs) > 0:
                    for orig_ix, correct_ix in track(
                        remap_dict.items(), description="uploading infer data..."
                    ):
                        val = outputs.pop(orig_ix)
                        logits_dict[correct_ix] = {}

                        for a in activation:
                            logits_dict[correct_ix][a] = val[a]
                            if correct_ix not in classification_distribution:
                                classification_distribution[correct_ix] = {}
                            if a in ["pre_softmax", "logits"]:
                                softmax_val = softmax(val[a])
                            if a == "softmax":
                                softmax_val = val[a]
                            classification_distribution[correct_ix][
                                self.current_loop
                            ] = {
                                "confidence": self.calculate_confidence(softmax_val),
                                "margin": self.calculate_margin(softmax_val),
                                "entropy": self.calculate_entropy(softmax_val),
                            }

                            # if (
                            #     "input_data" in val
                            #     or correct_ix in self._uploaded_infer_index
                            # ):
                            #     image_key = os.path.join(
                            #         self.expt_dir, "infer", str(correct_ix) + ".png"
                            #     )
                            #     if correct_ix not in self._uploaded_infer_index:
                            #         self.client.upload_object(
                            #             object=val["input_data"],
                            #             storage_bucket=self.bucket_name,
                            #             object_key=image_key,
                            #             format="png",
                            #         )
                            #         self._uploaded_infer_index.append(correct_ix)
                            # else:
                            #     image_key = None
                            # confidence_vals.append(
                            #     {
                            #         "index": correct_ix,
                            #         "prediction": val["prediction"],
                            #         "confidence": round(
                            #             softmax(val[a])[val["prediction"]], 2
                            #         ),
                            #         "current_loop": self.current_loop + 1,
                            #         "datamap_type": "infer_datamap",
                            #         "s3_image_path": image_key,
                            #         "project_id": self.project_id,
                            #         "experiment_id": self.experiment_id,
                            #     }
                            # )

                object_key = os.path.join(
                    self.expt_dir, "logits_{}.pkl".format(self.current_loop)
                )
                object_key = f"{self.expt_dir}/logits_{self.current_loop}.pkl"
                self.client.upload_object(
                    object=logits_dict,
                    storage_bucket=self.bucket_name,
                    object_key=object_key,
                    format="pkl",
                )
                self.client.upload_object(
                    object=classification_distribution,
                    storage_bucket=self.bucket_name,
                    object_key=f"{self.expt_dir}/confidence_logit_entropy.pkl",
                    format="pkl",
                )
            elif "object_detection" or "instance_segmentation" in self.task_type:
                object_key = os.path.join(
                    self.expt_dir, "infer_outputs_{}.db".format(self.current_loop)
                )
                logits_conn = create_connection(local_db)

                if logits_conn is not None:
                    logits_cur = logits_conn.cursor()
                    sql = """
                            UPDATE indexes
                            SET row_id = ?
                            WHERE id = ?
                    """
                    for index, correct_ix in enumerate(self.unlabeled):
                        logits_cur.execute(sql, (correct_ix, index))

                    logits_conn.close()
                else:
                    raise ConnectionError("Cannot connect to softmax database")

                # upload sqlite db
                self.client.upload_file(
                    storage_bucket=self.bucket_name,
                    file_path=local_db,
                    object_key=object_key,
                )
                # delete local version
                os.remove(local_db)
            elif "regression" in self.task_type:
                unlabelled_predictions = outputs["outputs"]

                object_key = os.path.join(
                    self.expt_dir,
                    "regression_predictions_{}.pkl".format(self.current_loop),
                )
                self.client.upload_object(
                    object=unlabelled_predictions,
                    storage_bucket=self.bucket_name,
                    object_key=object_key,
                    format="pkl",
                )
            else:
                raise ValueError("Use case not found")
        else:
            raise ValueError(
                "Infer outputs not returned. You must return outputs from the implemented infer() "
                "function. Please refer to the README on the Alectio SDK repository."
            )

        return

    def __call__(self):
        # SAVE DATAMAP SIZE
        datamap_size = len(self.getstate_fn(self.args).keys())
        self.console.info(f"TRAINING DATASET SIZE: {datamap_size}")
        error, response = self.api_client.POST_REQUEST(
            end_point="/sdk/v2/save_training_size",
            payload={"token": self.client_token, "train_dataset_size": datamap_size},
        )
        if error:
            self.console.error(
                "SOMETHING WENT WRONG WHILE SAVING YOUR TRAINING DATASET SIZE !\n PLEASE RECHECK"
            )
            return

        # START EXPERIMENT
        error, response = self.api_client.POST_REQUEST(
            end_point="/sdk/v2/start_experiment", payload={"token": self.client_token}
        )
        if error:
            self.console.error(
                "SOMETHING WENT WRONG WHILE STARTING THE EXPERIMENT !\n PLEASE CONTACT ALECTIO ADMIN"
            )
            return
        if response["status"] == "STARTED":
            count = 0
            while True:
                error, response = self.api_client.POST_REQUEST(
                    end_point="/sdk/v2/sdk_response",
                    payload={"token": self.client_token},
                )
                if error:
                    self.console.error(
                        "SOMETHING WENT WRONG WHILE TRYING TO CONNECT SDK WITH ALECITO SERVER !!\nREACH OUT TO ALECTIO ADMIN"
                    )
                    return
                # print("********")
                # print(response)
                if response["status"] == "FETCHED":
                    self.co2_tracker.start()
                    print("\n")
                    one_loop_response = self.one_loop(response)
                    count = 0
                    self.co2_tracker.stop()
                if response["status"] == "PAUSED":
                    self.console.warning(response["message"])
                    break
                if response["status"] == "FINISHED":
                    print("\n")
                    self.console.message("Experiment complete")

                    # co2 emissions upload to bucket and delete local file
                    local_emission_file_path_json = os.path.join(
                        self.co2_dir, "emissions.json"
                    )
                    local_emission_file_path_csv = os.path.join(
                        self.co2_dir, "emissions.csv"
                    )
                    co2_df = pd.read_csv(local_emission_file_path_csv)
                    co2_df.to_json(local_emission_file_path_json)
                    emission_file_path = os.path.join(
                        self.expt_dir, f"emissions_{self.experiment_id}.json"
                    )
                    self.client.upload_file(
                        storage_bucket=self.bucket_name,
                        file_path=local_emission_file_path_json,
                        object_key=emission_file_path,
                    )
                    os.remove(local_emission_file_path_csv)
                    os.remove(local_emission_file_path_json)

                    # time_saved upload to bucket and delete local file
                    local_time_saved_file_path_json = os.path.join(
                        self.time_saved_dir, "time_saved.json"
                    )
                    local_time_saved_file_path_csv = os.path.join(
                        self.time_saved_dir, self.time_saved_file
                    )
                    time_saved_df = pd.read_csv(local_time_saved_file_path_csv)
                    time_saved_df.to_json(local_time_saved_file_path_json)
                    time_saved_file_path = os.path.join(
                        self.expt_dir, f"time_saved_{self.experiment_id}.json"
                    )
                    self.client.upload_file(
                        storage_bucket=self.bucket_name,
                        file_path=local_time_saved_file_path_json,
                        object_key=time_saved_file_path,
                    )
                    os.remove(local_time_saved_file_path_json)
                    os.remove(local_time_saved_file_path_csv)
                    break
                if response["status"] == "WAITING":
                    if count == 0:
                        print("\n")

                        self.console.warning("Waiting for server.")
                        count += 1
                        time.sleep(10)
                    else:
                        time.sleep(10)
                        print(".", end="", flush=True)
                if response["status"] == "FAILED":
                    self.console.warning("YOUR EXPERIMENT IS FAILED.")
                    break
        elif response["status"] == "PAUSED":
            self.console.warning(response["message"])

        elif response["status"] == "FINISHED":
            self.console.message(response["message"])



from __future__ import division
import math
import time
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import numpy as np


def bbox_iou(box1, box2, x1y1x2y2=True):
    """
    Returns the IoU of two bounding boxes
    """
    if not x1y1x2y2:
        # Transform from center and width to exact coordinates
        b1_x1, b1_x2 = box1[:, 0] - box1[:, 2] / 2, box1[:, 0] + box1[:, 2] / 2
        b1_y1, b1_y2 = box1[:, 1] - box1[:, 3] / 2, box1[:, 1] + box1[:, 3] / 2
        b2_x1, b2_x2 = box2[:, 0] - box2[:, 2] / 2, box2[:, 0] + box2[:, 2] / 2
        b2_y1, b2_y2 = box2[:, 1] - box2[:, 3] / 2, box2[:, 1] + box2[:, 3] / 2
    else:
        # Get the coordinates of bounding boxes
        b1_x1, b1_y1, b1_x2, b1_y2 = box1[:, 0], box1[:, 1], box1[:, 2], box1[:, 3]
        b2_x1, b2_y1, b2_x2, b2_y2 = box2[:, 0], box2[:, 1], box2[:, 2], box2[:, 3]

    # get the corrdinates of the intersection rectangle
    inter_rect_x1 = torch.max(b1_x1, b2_x1)
    inter_rect_y1 = torch.max(b1_y1, b2_y1)
    inter_rect_x2 = torch.min(b1_x2, b2_x2)
    inter_rect_y2 = torch.min(b1_y2, b2_y2)
    # Intersection area
    inter_area = torch.clamp(inter_rect_x2 - inter_rect_x1 + 1, min=0) * torch.clamp(
        inter_rect_y2 - inter_rect_y1 + 1, min=0
    )
    # Union Area
    b1_area = (b1_x2 - b1_x1 + 1) * (b1_y2 - b1_y1 + 1)
    b2_area = (b2_x2 - b2_x1 + 1) * (b2_y2 - b2_y1 + 1)

    iou = inter_area / (b1_area + b2_area - inter_area + 1e-16)

    return iou


def non_max_suppression(prediction, num_classes, conf_thres=0.5, nms_thres=0.4):
    """
    Removes detections with lower object confidence score than 'conf_thres' and performs
    Non-Maximum Suppression to further filter detections.
    Returns detections with shape:
        (x1, y1, x2, y2, object_conf, class_score, class_pred)
    """

    # From (center x, center y, width, height) to (x1, y1, x2, y2)
    box_corner = prediction.new(prediction.shape)
    box_corner[:, :, 0] = prediction[:, :, 0] - prediction[:, :, 2] / 2
    box_corner[:, :, 1] = prediction[:, :, 1] - prediction[:, :, 3] / 2
    box_corner[:, :, 2] = prediction[:, :, 0] + prediction[:, :, 2] / 2
    box_corner[:, :, 3] = prediction[:, :, 1] + prediction[:, :, 3] / 2
    prediction[:, :, :4] = box_corner[:, :, :4]

    output = [None for _ in range(len(prediction))]
    for image_i, image_pred in enumerate(prediction):
        # Filter out confidence scores below threshold
        conf_mask = (image_pred[:, 4] >= conf_thres).squeeze()
        image_pred = image_pred[conf_mask]
        # If none are remaining => process next image
        if not image_pred.size(0):
            continue
        # Get score and class with highest confidence
        class_conf, class_pred = torch.max(
            image_pred[:, 5 : 5 + num_classes], 1, keepdim=True
        )
        # Detections ordered as (x1, y1, x2, y2, obj_conf, class_conf, class_pred)
        detections = torch.cat(
            (image_pred[:, :5], class_conf.float(), class_pred.float()), 1
        )
        # Iterate through all predicted classes
        unique_labels = detections[:, -1].cpu().unique()
        if prediction.is_cuda:
            unique_labels = unique_labels.cuda()
        for c in unique_labels:
            # Get the detections with the particular class
            detections_class = detections[detections[:, -1] == c]
            # Sort the detections by maximum objectness confidence
            _, conf_sort_index = torch.sort(detections_class[:, 4], descending=True)
            detections_class = detections_class[conf_sort_index]
            # Perform non-maximum suppression
            max_detections = []
            while detections_class.size(0):
                # Get detection with highest confidence and save as max detection
                max_detections.append(detections_class[0].unsqueeze(0))
                # Stop if we're at the last detection
                if len(detections_class) == 1:
                    break
                # Get the IOUs for all boxes with lower confidence
                ious = bbox_iou(max_detections[-1], detections_class[1:])
                # Remove detections with IoU >= NMS threshold
                detections_class = detections_class[1:][ious < nms_thres]

            max_detections = torch.cat(max_detections).data
            # Add max detections to outputs
            output[image_i] = (
                max_detections
                if output[image_i] is None
                else torch.cat((output[image_i], max_detections))
            )

    return output



import warnings
from copy import deepcopy
from typing import Dict, List, Set, Union

import numpy as np



class DuplicatedElementWarning(UserWarning):
    """If contains repeated elements."""

    ...


class InexistentElementWarning(UserWarning):
    """If the element selected does not exist."""

    ...


class OverwritedValueWarning(UserWarning):
    """If a user could accidentally be overwriting a previously stored value."""

    ...



class DataPool:
    """A basic container for unlabeled and labeled data."""

    def __init__(
        self, unlabeled: Union[List[int], np.ndarray, Set[int]], *args, **kwargs
    ) -> None:
        """Create a new instance of DataPool.

        Args:
            unlabeled (Union[List[int], np.ndarray, Set[int]]): List, numpy array, or set of distinct unlabeled indices.
        """  # noqa: E501

        self._labeled = set()
        self._unlabeled = set(unlabeled)
        self.loop_dictionary: Dict[int, Dict[int, set]] = {}

    @property
    def labeled(self) -> List[int]:
        """Getter method for labeled attribute.

        Returns:
            List[int]: Return labeled indices as a/an (unordered) list
        """
        return list(self._labeled)

    @property
    def unlabeled(self) -> List[int]:
        """Getter method for unlabeled attribute.

        Returns:
            List[int]: Return unlabeled indices as a/an (unordered) list
        """
        return list(self._unlabeled)

    def __contains__(self, index: int) -> bool:
        """Check whether an index presents in DataPool or not.

        Args:
            index (int): index to test on

        Returns:
            bool: True if the index presents in labeled or unlabeled pool, False otherwise
        """  # noqa: E501
        return index in self.labeled or index in self.unlabeled

    def _input_sanity_check(
        self, container: str, value: int, labeled: bool = False
    ) -> int:
        """Sanity check for the value to be processed.

        Args:
            container (str): DEPRECATED. Name of the pool to be checked, must be "labeled" or "unlabeled".
            value (int): value that needs to run checks on
            labeled (bool, optional): Whether to check the value on the labeled or unlabeled pool. Defaults to False.
        """  # noqa: E501
        assert container in [
            "labeled",
            "unlabeled",
        ], "Invalid container value, must be 'labeled' or 'unlabeled'"
        warnings.warn(
            "Usage of 'container' will be removed in the future. "
            "Please use 'labeled' argument instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if not isinstance(value, int):
            warnings.warn(
                "Value data type must be int. Trying to convert it now...",
                RuntimeWarning,
                stacklevel=3,
            )
            return int(value)
        else:
            return value

    def add(self, container: str, value: int, labeled: bool = False) -> None:
        """Add an index to the labeled or unlabeled pool.

        Args:
            container (str): DEPRECATED. Name of the pool to be added, must be "labeled" or "unlabeled".
            value (int): A new index to be added
            labeled (bool, optional): Whether to add a new index to the labeled or unlabeled pool. Defaults to False.
        """  # noqa: E501
        value = self._input_sanity_check(container, value, labeled)
        pool = self._labeled if container == "labeled" or labeled else self._unlabeled
        if value in pool:
            warnings.warn(
                f"Element {value} is already in the pool, skipping...",
                category=DuplicatedElementWarning,
                stacklevel=2,
            )
        else:
            pool.add(value)

    def delete(self, container: str, value: int, labeled: bool = False) -> None:
        """Discard an index in the labeled or unlabeled pool.

        Args:
            container (str): DEPRECATED. Name of the pool to be removed from, must be "labeled" or "unlabeled".
            value (int): A new index to be deleted
            labeled (bool, optional): Whether to remove the index in the labeled or unlabeled pool. Defaults to False.
        """  # noqa: E501
        value = self._input_sanity_check(container, value, labeled)
        pool = self._labeled if container == "labeled" or labeled else self._unlabeled
        if value not in pool:
            warnings.warn(
                f"Element {value} to discard is not in the pool, skipping...",
                category=InexistentElementWarning,
                stacklevel=2,
            )
        else:
            pool.remove(value)

    def update(
        self,
        container: str,
        values: Union[List[int], np.ndarray, Set[int]],
        labeled: bool = False,
    ) -> None:
        """Add multiple indices to the labeled or unlabeled pool.

        Args:
            container (str): DEPRECATED. Name of the pool to be added, must be "labeled" or "unlabeled".
            values (Union[List[int], np.ndarray, Set[int]]): indices to be added. Duplicated indices will be discarded.
            labeled (bool, optional): Whether to add indices to the labeled or unlabeled pool. Defaults to False.
        """  # noqa: E501
        _ = self._input_sanity_check(
            container, -1, labeled
        )  # TODO: properly handle the -1 here
        pool = self._labeled if container == "labeled" or labeled else self._unlabeled
        pool |= set(values)

    def difference_update(
        self,
        container: str,
        values: Union[List[int], np.ndarray, Set[int]],
        labeled: bool = False,
    ) -> None:
        """DEPRECATED. Discard multiple indices in the labeled or unlabeled pool.

        Args:
            container (str): DEPRECATED. Name of the pool to be removed from, must be "labeled" or "unlabeled".
            values ((Union[List[int], np.ndarray, Set[int]])): indices to be removed. Duplicated indices will be discarded.
            labeled (bool, optional): Whether to remove indices in the labeled or unlabeled pool. Defaults to False.
        """  # noqa: E501
        warnings.warn(
            "Usage of 'difference_update' will be remove in the future. "
            "Please use 'difference' method instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.difference(container, values, labeled)

    def difference(
        self,
        container: str,
        values: Union[List[int], np.ndarray, Set[int]],
        labeled: bool = False,
    ) -> None:
        """Discard multiple indices in the labeled or unlabeled pool.

        Args:
            container (str): DEPRECATED. Name of the pool to be removed from, must be "labeled" or "unlabeled".
            values ((Union[List[int], np.ndarray, Set[int]])): indices to be removed. Duplicated indices will be discarded.
            labeled (bool, optional): Whether to remove indices in the labeled or unlabeled pool. Defaults to False.
        """  # noqa: E501
        _ = self._input_sanity_check(
            container, -1, labeled
        )  # TODO: properly handle the -1 here
        pool = self._labeled if container == "labeled" or labeled else self._unlabeled
        pool -= set(values)

    def loop_snapshot(self, loop_number: int, verbose=False) -> None:
        """Store the current state of the labeled and unlabeled pool as a dict. It will enable you
        to call on self.loop_dictionary to access the unlabeled or labeled pool at a particular
        step.

        Args:
            loop_number (int): An integer that represents the loop number.
            verbose (bool, optional): Print logging details. Defaults to False.
        """  # noqa: E501
        if not isinstance(loop_number, int):
            warnings.warn(
                "Value data type must be int. Trying to convert it now...",
                RuntimeWarning,
                stacklevel=3,
            )
            loop_number = int(loop_number)

        self.loop_dictionary[loop_number] = {
            "unlabeled": deepcopy(self.unlabeled),
            "labeled": deepcopy(self.labeled),
        }
        if verbose:
            print(f"Loop snapshot at loop {loop_number} successful")

    def random_sample(
        self, container: str, quantity: int, seed: int = None, labeled: bool = False
    ) -> np.ndarray:
        """Randomly sample indices from the labeled or unlabeled pool.

        Args:
            container (str): DEPRECATED. Name of the pool to sample, must be "labeled" or "unlabeled".
            quantity (int): Number of items to sample.
            seed (int, optional): Seed value to be used in the random process. Defaults to None.
            labeled (bool, optional): Whether to sample in the labeled or unlabeled pool. Defaults to False.

        Returns:
            np.ndarray: Selected indices that were randomly sampled.
        """  # noqa: E501
        # TODO: error checking on quantity
        quantity = self._input_sanity_check(container, quantity, labeled)
        pool = self.labeled if container == "labeled" or labeled else self.unlabeled
        selected = np.random.default_rng(seed).choice(pool, quantity, replace=False)
        return selected



import numpy as np
from sklearn.cluster import KMeans
from sklearn.manifold import TSNE
import random
from typing import List


def Hbeta(D=np.array([]), beta=1.0):
    """
        Compute the perplexity and the P-row for a specific value of the
        precision of a Gaussian distribution.
    """

    # Compute P-row and corresponding perplexity
    P = np.exp(-D.copy() * beta)
    sumP = sum(P)
    H = np.log(sumP) + beta * np.sum(D * P) / sumP
    P = P / sumP
    return H, P


def x2p(X=np.array([]), tol=1e-5, perplexity=30.0):
    """
        Performs a binary search to get P-values in such a way that each
        conditional Gaussian has the same perplexity.
    """

    # Initialize some variables
    print("Computing pairwise distances...")
    (n, d) = X.shape
    sum_X = np.sum(np.square(X), 1)
    D = np.add(np.add(-2 * np.dot(X, X.T), sum_X).T, sum_X)
    P = np.zeros((n, n))
    beta = np.ones((n, 1))
    logU = np.log(perplexity)

    # Loop over all datapoints
    for i in range(n):

        # Print progress
        if i % 500 == 0:
            print("Computing P-values for point %d of %d..." % (i, n))

        # Compute the Gaussian kernel and entropy for the current precision
        betamin = -np.inf
        betamax = np.inf
        Di = D[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))]
        (H, thisP) = Hbeta(Di, beta[i])

        # Evaluate whether the perplexity is within tolerance
        Hdiff = H - logU
        tries = 0
        while np.abs(Hdiff) > tol and tries < 50:

            # If not, increase or decrease precision
            if Hdiff > 0:
                betamin = beta[i].copy()
                if betamax == np.inf or betamax == -np.inf:
                    beta[i] = beta[i] * 2.
                else:
                    beta[i] = (beta[i] + betamax) / 2.
            else:
                betamax = beta[i].copy()
                if betamin == np.inf or betamin == -np.inf:
                    beta[i] = beta[i] / 2.
                else:
                    beta[i] = (beta[i] + betamin) / 2.

            # Recompute the values
            (H, thisP) = Hbeta(Di, beta[i])
            Hdiff = H - logU
            tries += 1

        # Set the final row of P
        P[i, np.concatenate((np.r_[0:i], np.r_[i+1:n]))] = thisP

    # Return final P-matrix
    print("Mean value of sigma: %f" % np.mean(np.sqrt(1 / beta)))
    return P


def pca(X=np.array([]), no_dims=50):
    """
        Runs PCA on the NxD array X in order to reduce its dimensionality to
        no_dims dimensions.
    """

    print("Preprocessing the data using PCA...")
    (n, d) = X.shape
    X = X - np.tile(np.mean(X, 0), (n, 1))
    (l, M) = np.linalg.eig(np.dot(X.T, X))
    Y = np.dot(X, M[:, 0:no_dims])
    return Y


def tsne(X=np.array([]), no_dims=2, initial_dims=50, perplexity=30.0):
    """
        Runs t-SNE on the dataset in the NxD array X to reduce its
        dimensionality to no_dims dimensions. The syntaxis of the function is
        `Y = tsne(X, no_dims, initial_dim, perplexity), where X is an NxD NumPy array.
    """

    # Check inputs
    if isinstance(no_dims, float):
        print("Error: array X should have type float.")
        return -1
    if round(no_dims) != no_dims:
        print("Error: number of dimensions should be an integer.")
        return -1

    # Initialize variables
    X = pca(X, initial_dims).real
    (n, d) = X.shape
    max_iter = 500
    initial_momentum = 0.5
    final_momentum = 0.8
    eta = 500
    min_gain = 0.01
    Y = np.random.randn(n, no_dims)
    dY = np.zeros((n, no_dims))
    iY = np.zeros((n, no_dims))
    gains = np.ones((n, no_dims))

    # Compute P-values
    P = x2p(X, 1e-5, perplexity)
    P = P + np.transpose(P)
    P = P / np.sum(P)
    P = P * 4.									# early exaggeration
    P = np.maximum(P, 1e-12)

    # Run iterations
    for iter in range(max_iter):

        # Compute pairwise affinities
        sum_Y = np.sum(np.square(Y), 1)
        num = -2. * np.dot(Y, Y.T)
        num = 1. / (1. + np.add(np.add(num, sum_Y).T, sum_Y))
        num[range(n), range(n)] = 0.
        Q = num / np.sum(num)
        Q = np.maximum(Q, 1e-12)

        # Compute gradient
        PQ = P - Q
        for i in range(n):
            dY[i, :] = np.sum(np.tile(PQ[:, i] * num[:, i], (no_dims, 1)).T * (Y[i, :] - Y), 0)

        # Perform the update
        if iter < 20:
            momentum = initial_momentum
        else:
            momentum = final_momentum
        gains = (gains + 0.2) * ((dY > 0.) != (iY > 0.)) + \
                (gains * 0.8) * ((dY > 0.) == (iY > 0.))
        gains[gains < min_gain] = min_gain
        iY = momentum * iY - eta * (gains * dY)
        Y = Y + iY
        Y = Y - np.tile(np.mean(Y, 0), (n, 1))

        # Compute current value of cost function
        if (iter + 1) % 100 == 0:
            C = np.sum(P * np.log(P / Q))
            print("Iteration %d: error is %f" % (iter + 1, C))

        # Stop lying about P-values
        if iter == 100:
            P = P / 4.

    # Return solution
    return Y


# Automatic ideal cluster detection
def deter_clusters(X, rng):
    WCSS = []
    for i in range(1, rng):
        kmeans = KMeans(i)
        kmeans.fit(X)
        WCSS.append(kmeans.inertia_)

    secondDerivative = [0 for i in range(len(WCSS))]
    for i in range(1, len(WCSS)-1):
        secondDerivative[i] = WCSS[i+1] + WCSS[i-1] - 2 * WCSS[i]
    
    optimal_clusters = secondDerivative.index(max(secondDerivative))+2
    return optimal_clusters, secondDerivative, WCSS


# Perform KMeans clustering on the data
def cluster_data(X, clusters):
    num_clusters = clusters
    kmeans = KMeans(num_clusters)
    kmeans.fit(X)
    labels = list(kmeans.labels_)
    return labels


# Simple Stratification
def stratified_selection(labels: List):
    '''
    Simple stratified selection.

    Args:
    labels: List()    # List of labels

    Returns:

    selected: List()  # List of selected indices
    '''
    selection_pool = dict()
    for i in range(len(labels)):
        if labels[i] not in list(selection_pool.keys()):
            selection_pool[labels[i]] = []
        selection_pool[labels[i]].append(i)

    selected = []

    for key in selection_pool.keys():
        num_samples = int(len(selection_pool[key])*.1)
        samples = random.sample(selection_pool[key], num_samples)
        selected.extend(samples)
    
    return selected


# Clustering with stratified selection
def stratified_selection_from_clusters(X: np.ndarray, clusters: int, final_features: int = 2, start_features: int = 50, perplexity: float = 30.0, sklearn: bool = True):
    '''
    Stratified selection using clustering.

    Args:
    X: np.ndarray       # Data
    clusters: int       # Number of clusters
    final_features: int # Number of features that will be used for clustering
    start_features: int # Features to be extracted from PCA
    perplexity: float   # The perplexity is related to the number of nearest neighbors that
                        is used in other manifold learning algorithms. Larger datasets
                        usually require a larger perplexity. Consider selecting a value
                        between 5 and 50. Different values can result in significantly
                        different results. The perplexity must be less than the number
                        of samples.
    sklearn: bool       # Use sklearn implementation True/False

    Returns:
    selected: List()    # List of selected indices
    '''
    # Run TSNE
    if sklearn:
        # Sklearn Implementation
        Y = TSNE(n_components=final_features, learning_rate='auto', perplexity=perplexity).fit_transform(X)
    else:
        # Orignal Implementation
        Y = tsne(X, final_features, start_features, perplexity)
    # Supervised learning 
    labels = cluster_data(Y, clusters)
    # Select indices
    selected = stratified_selection(labels)

    return selected


# Auto Clustering with stratified selection
def stratified_selection_from_auto_clusters(X: np.ndarray, cluster_range: int, final_features: int = 2, start_features: int = 50, perplexity: float = 30.0, sklearn: bool = True):
    '''
    Stratified selection using auto clustering. Here the ideal clusters are determined automatically.
    The user needs to specify the range in which he wants to find ideal clusters. 
    The ideal number of clusters are determined using the elbow method.

    Args:
    X: np.ndarray       # Data
    clusters_range: int # Number of clusters
    final_features: int # Number of features that will be used for clustering
    start_features: int # Features to be extracted from PCA
    perplexity: float   # The perplexity is related to the number of nearest neighbors that
                        is used in other manifold learning algorithms. Larger datasets
                        usually require a larger perplexity. Consider selecting a value
                        between 5 and 50. Different values can result in significantly
                        different results. The perplexity must be less than the number
                        of samples.
    sklearn: bool       # Use sklearn implementation True/False

    Returns:
    selected: List()    # List of selected indices
    '''
    
    # Run TSNE
    if sklearn:
        # Sklearn Implementation
        Y = TSNE(n_components=final_features, learning_rate='auto', perplexity=perplexity).fit_transform(X)
    else:
        # Orignal Implementation
        Y = tsne(X, final_features, start_features, perplexity)
    # Unsupervised learning 
    clusters, secondDerivative, WCSS = deter_clusters(Y, cluster_range)
    labels = cluster_data(Y, clusters)
    # Select indices
    selected = stratified_selection(labels)

    return selected


import numpy as np


class RegressionMetrics(object):
    """
    Description:

        MeanAbsoluteError(MAE)
        MeanSquareError(MSE)
        RootMeanSquareError(RMSE)
        RootMeanSquareLogError(RMSLE)
        R-Squared(R2)
        Adjusted_R2

    Evaluation Metrics for Time-Series:

    Args:
        target: list
        list of true numerical values/number that has to be predicted

        predicted: list
        list of values prediction corresponding to the target values

        k:
        integer representing the number of features used for prediction in regression
    """

    def __init__(self, target, prediction, k=1):
        self.target = target
        self.prediction = prediction
        self.k = k
        self._compute_metrics(target, prediction)

    def _compute_metrics(self, target, prediction, k=1):
        self.n = len(prediction)
        target, prediction = np.array(target), np.array(prediction)
        self.mean = np.mean(target)

        self.absolute_diff = np.abs(target - prediction)
        self.TSS = np.sum(np.square(prediction - self.mean))
        self.RSS = np.sum(np.square(prediction - target))

        self.MAE = np.mean(self.absolute_diff)
        self.MSE = np.mean(np.square(self.absolute_diff))
        self.RMSE = np.sqrt(self.MSE)
        self.RMSLE = np.log(self.RMSE)
        self.R2 = 1 - self.RSS / self.TSS
        self.adjusted_R2 = 1 - ((self.n - 1) / self.n - k - 1) * (1 - self.R2)

    def get_MAE(self):
        """
        Returns:
         Mean Absolute Error as float
        """
        return self.MAE

    def get_MSE(self):
        """
        Returns:
         Mean Square Error as float
        """
        return self.MSE

    def get_RSME(self):
        """
        Returns:
         Root Mean Square Error as float
        """
        return self.RSME

    def get_RMSLE(self):
        """
        Returns:
         Root Mean Square Log Error as float
        """
        return self.RMSLE

    def get_R2(self):
        """
        Returns:
         R-Square value as float between optimum (0-1)
        """
        return self.R2

    def get_adjusted_R2(self):
        """
        Returns:
         Adjusted R-Square value as float between (0-1)
        """
        return self.adjusted_R2



import os
import logging
import tensorflow as tf
import numpy as np



def settfreproduceability(seed= 42):
    """
        Function enables reproduceablity for your experiments so that you can compare easily
        Parameters:
        seed (int): seed value for global randomness
    """
    logging.warning("*** Modules are set to be deterministic , randomness in your modules will be avoided , Current seed value is {} change seed value if you want to try a different seed of parameters***".format(seed))
    #Pythonic determinism
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    #Pythonic determinism
    tf.random.set_seed(seed)
    os.environ['TF_DETERMINISTIC_OPS'] = '1'
    os.environ['TF_CUDNN_DETERMINISTIC'] = '1'
    # https://www.tensorflow.org/api_docs/python/tf/config/threading/set_inter_op_parallelism_threads
    tf.config.threading.set_inter_op_parallelism_threads(1)
    tf.config.threading.set_intra_op_parallelism_threads(1)
    from tfdeterminism import patch
    patch()


# -*- coding: utf-8 -*-
"""
Created on Sat May 16 12:50:47 2020

@author: arun
"""

import numpy as np


def compute_iou(detBoxes, GTBoxes):
    """
    Description:
        Computes IoU between Detected boxes and Ground truth boxes.
    Arguments:
        detBoxes : ndarray
            (N, 4) shaped array with detected bboxes
        GTBoxes : ndarray
            (M, 4) shaped array with Ground truth bboxes
        Returns: 
            (N, M) shaped numpy array with IoUs
    """
    area = (GTBoxes[:, 2] - GTBoxes[:, 0]) * (GTBoxes[:, 3] - GTBoxes[:, 1])

    iw = np.minimum(np.expand_dims(detBoxes[:, 2], axis=1), GTBoxes[:, 2]) - np.maximum(
        np.expand_dims(detBoxes[:, 0], 1), GTBoxes[:, 0]
    )
    ih = np.minimum(np.expand_dims(detBoxes[:, 3], axis=1), GTBoxes[:, 3]) - np.maximum(
        np.expand_dims(detBoxes[:, 1], 1), GTBoxes[:, 1]
    )

    iw = np.maximum(iw, 0)
    ih = np.maximum(ih, 0)

    ua = (
        np.expand_dims(
            (detBoxes[:, 2] - detBoxes[:, 0]) * (detBoxes[:, 3] - detBoxes[:, 1]),
            axis=1,
        )
        + area
        - iw * ih
    )

    ua = np.maximum(ua, np.finfo(float).eps)

    intersection = iw * ih

    return intersection / ua


def compute_ap(recall, precision):
    """ 
    Description:
        Compute the average precision, given the recall and precision curves.
    Arguments:
        recall:    The recall curve (list).
        precision: The precision curve (list).
    Returns:
        Average precision
    """
    mrec = np.concatenate(([0.0], recall, [1.0]))
    mpre = np.concatenate(([0.0], precision, [0.0]))

    # compute the precision envelope
    for i in range(mpre.size - 1, 0, -1):
        mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

    # To calculate area under PR curve, look for points where X axis (recall) changes value
    i = np.where(mrec[1:] != mrec[:-1])[0]

    # and sum (\Delta recall) * prec
    ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap



import os
import torch
import numpy as np
import logging
import random

def setpytorchreproduceability(seed = 42):
    """
        Function enables reproduceablity for your experiments so that you can compare easily
        Parameters:
        seed (int): seed value for global randomness
    """

    logging.warning("*** Modules are set to be deterministic , randomness in your modules will be avoided , Current seed value is {} change seed value if you want to try a different seed of parameters***".format(seed))
    #Pythonic determinism
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    #Pythonic determinism
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True


from sklearn.metrics import (
    f1_score,
    precision_score,
    recall_score,
    accuracy_score,
    confusion_matrix,
)
import numpy as np


class ClassificationMetrics(object):
    r"""
    Evaluation Metrics for classification

    Args:
        predictions: list,
            list of classes predicted for each record

        ground_truth: list,
            list of ground_truth classes for each record
    """

    def __init__(self, predictions, ground_truth):
        self.predictions = predictions
        self.ground_truth = ground_truth
        self._compute_metrics(predictions, ground_truth)

    def _compute_metrics(self, predictions, ground_truth):

        f1_scores = f1_score(ground_truth, predictions, average=None, zero_division=0)
        recall = recall_score(ground_truth, predictions, average=None, zero_division=0)
        precision = precision_score(
            ground_truth, predictions, average=None, zero_division=0
        )

        self.acc_per_class = {
            i: val
            for i, val in enumerate(
                confusion_matrix(ground_truth, predictions, normalize="true").diagonal()
            )
        }
        self.f1_score_per_class = {i: val for i, val in enumerate(f1_scores)}
        self.precision_per_class = {i: val for i, val in enumerate(precision)}
        self.recall_per_class = {i: val for i, val in enumerate(recall)}

        self.accuracy = accuracy_score(ground_truth, predictions)
        self.f1_score = f1_scores.mean()
        self.recall = recall.mean()
        self.precision = precision.mean()

        self.confusion_matrix = confusion_matrix(ground_truth, predictions)
        self.FP = self.confusion_matrix.sum(axis=0) - np.diag(self.confusion_matrix)
        self.FN = self.confusion_matrix.sum(axis=1) - np.diag(self.confusion_matrix)
        self.TP = self.confusion_matrix.diagonal()
        self.TN = self.confusion_matrix.sum() - (self.FP + self.FN + self.TP)
        self.label_disagreement = {
            k: v.round(3) for k, v in enumerate(self.FP / (self.FP + self.TN))
        }

        # calculate class wise confusion matrix
        all_classes = list(set(ground_truth))
        self.confusion_matrix_per_class = {}
        for class_label in all_classes:
            self.confusion_matrix_per_class[
                class_label
            ] = self.calculate_confusion_matrix_per_class(class_label)

        # converting numpy confusion matrix to list
        self.confusion_matrix = self.confusion_matrix.tolist()

    def calculate_confusion_matrix_per_class(self, class_of_interest):

        class_indices = [class_of_interest]

        tp = self.confusion_matrix[class_indices, class_indices][0]
        fn = self.confusion_matrix[class_indices, :].sum() - tp
        fp = self.confusion_matrix[:, class_indices].sum() - tp
        tn = self.confusion_matrix.sum() - tp - fn - fp

        return [[tp, fn], [fp, tn]]

    def get_accuracy(self):
        r"""
        Returns:
            accuracy as a float
        """
        return self.accuracy

    def get_f1_score(self):
        r"""
        Returns:
            F1 score as a float
        """
        return self.f1_score

    def get_f1_score_per_class(self):
        r"""
        Returns:
            F1 score per class as a dict
        """
        return self.f1_score_per_class

    def get_precision_per_class(self):
        r"""
        Returns:
            precision per class as a dict
        """
        return self.precision_per_class

    def get_precision(self):
        r"""
        Returns:
            precision as a float
        """
        return self.precision

    def get_recall_per_class(self):
        r"""
        Returns:
            recall per class as a dict
        """
        return self.recall_per_class

    def get_recall(self):
        r"""
        Returns:
            recall as a float
        """
        return self.recall

    def get_confusion_matrix(self):
        r"""
        Returns:
            confusion matrix as a list
        """
        return self.confusion_matrix

    def get_acc_per_class(self):
        r"""
        Returns:
            acc_per_class as a dict
        """
        return self.acc_per_class

    def get_label_disagreement(self):
        r"""
        Returns:
            label disagreement as a dict
        """
        return self.label_disagreement

    def get_confusion_matrix_per_class(self):
        return self.confusion_matrix_per_class



import linecache
import os
import yaml
import random
import tracemalloc
import sqlite3
import numpy as np
from sqlite3 import Error


def create_database(filepath):
    database = filepath
    sql_create_indexes_table = """ CREATE TABLE IF NOT EXISTS indexes (
                                    id integer PRIMARY KEY,
                                    row_id integer NOT NULL,
                                    selected_index text NOT NULL
                                ); """

    conn = create_connection(database)
    if conn is not None:
        create_table(conn, sql_create_indexes_table)
    else:
        print("Error! cannot create the database connection.")
    return conn


def create_connection(db_file):
    """ create a database connection to a SQLite database """
    conn = None
    try:
        conn = sqlite3.connect(db_file)
        return conn
    except Error as e:
        print(e)


def create_table(conn, create_table_sql):
    """ create a table from the create_table_sql statement
    :param conn: Connection object
    :param create_table_sql: a CREATE TABLE statement
    :return:
    """
    try:
        c = conn.cursor()
        c.execute(create_table_sql)
    except Error as e:
        print(e)


def add_index(conn, row_id, array):
    """
    Convert numpy array to bytes and insert into database
    :param conn: SQLite Connection Object
    :param array: A numpy array
    :return: id of last row entered
    """
    asbyte = array.dumps()
    sql = ''' INSERT INTO indexes(row_id, selected_index)
              VALUES(?, ?) '''
    cur = conn.cursor()
    cur.execute(sql, (row_id, asbyte))
    conn.commit()
    return cur.lastrowid





import os
from typing import Dict
import json
from google.cloud import storage
from google.oauth2 import service_account
import requests 

import crcmod
import six
from .api_client import APIClient


from google.cloud import kms

# READ SDK CONFIG FILE
dir_path = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(dir_path, "config.json"), "r") as f:
    config = json.load(f)

def seed_comparison(best_seed_info:Dict,new_seed_info:Dict):
    return new_seed_info if new_seed_info['metric'] > best_seed_info['metric'] else best_seed_info


def crc32c(data):
    """
    Calculates the CRC32C checksum of the provided data.

    Args:
        data: the bytes over which the checksum should be calculated.

    Returns:
        An int representing the CRC32C checksum of the provided bytes.
    """
    crc32c_fun = crcmod.predefined.mkPredefinedCrcFun('crc-32c')
    return crc32c_fun(six.ensure_binary(data))


def decrypt_symmetric(project_id, location_id, key_ring_id, key_id, ciphertext,kms_cred_info):
    """
    Decrypt the ciphertext using the symmetric key

    Args:
        project_id (string): Google Cloud project ID (e.g. 'my-project').
        location_id (string): Cloud KMS location (e.g. 'us-east1').
        key_ring_id (string): ID of the Cloud KMS key ring (e.g. 'my-key-ring').
        key_id (string): ID of the key to use (e.g. 'my-key').
        ciphertext (bytes): Encrypted bytes to decrypt.
        kms_cred_info (dict) : Decryption keys 

    Returns:
        DecryptResponse: Response including plaintext.

    """
    

    # Import the client library.

    # with open("kms-test.json") as info:
    #     info = json.load(info)
   
    kms_cred_info = json.loads(kms_cred_info)
    
    kms__local_credentials = service_account.Credentials.from_service_account_info(kms_cred_info)
    

    # Create the client.
    client = kms.KeyManagementServiceClient(credentials=kms__local_credentials)
    # Build the key name.
    key_name = client.crypto_key_path(project_id, location_id, key_ring_id, key_id)

    # Optional, but recommended: compute ciphertext's CRC32C.
    # See crc32c() function defined below.
    ciphertext_crc32c = crc32c(ciphertext)

    # Call the API.
    decrypt_response = client.decrypt(
        request={'name': key_name, 'ciphertext': ciphertext, 'ciphertext_crc32c': ciphertext_crc32c})

    # Optional, but recommended: perform integrity verification on decrypt_response.
    # For more details on ensuring E2E in-transit integrity to and from Cloud KMS visit:
    # https://cloud.google.com/kms/docs/data-integrity-guidelines
    if not decrypt_response.plaintext_crc32c == crc32c(decrypt_response.plaintext):
        raise Exception('The response received from the server was corrupted in-transit.')
    # End integrity verification

    return decrypt_response.plaintext






 
def get_kms_credentials(exp_token):
    url = f"{config['backend_ip']}/api/v2/get_decryption_keys/experiment_token/{exp_token}"
    response_credentials = getRequest(url)
    if response_credentials:
        return response_credentials
    
    return None
    
def get_storage_credentials(exp_token):
   
    url = f"{config['backend_ip']}/api/v2/get_credentials/experiment_token/{exp_token}"
    response = getRequest(url)
    
    if response:

        #response = response.json()
        data = bytes(response['data'],encoding='latin-1')
        kms_creds = get_kms_credentials(exp_token=exp_token)
       


        decrypted_storage_cred=decrypt_symmetric("dev-backend-369517","global","usercred-ring","symmetric-key",data,kms_creds['data'])
       
        return str(decrypted_storage_cred,'UTF-8')
    return None

def getRequest(url):
    response = requests.get(url, headers={"X-Alectio-Flavor": "PRO"})
    if response.status_code  in [200, 201, 202, 203, 204, 2005, 206]:
        return response.json()        
    return None


import numpy as np


class TimeSeriesMetrics(object):
    """
    Description:

        Four different categories of Error Metrics for Time-Series:

          Scale-Dependent Metrics:(+Regression)
              MeanAbsoluteError(MAE), MeanSquareError(MSE), RootMeanSquareError(RMSE), RootMeanSquareLogError(RMSLE), R-Squared(R2)

          Percentage-Error Metrics:
              MeanAbsolutePercentageError(MAPE), SymmetricMeanAbsolutePrecentageError(sMAPE)

          Relative-Error Metrics:(benchmark required)
              MedianRelativeAbsoluteError(MdRAE), GeometricMeanAbsoluteError(GMRAE)

          Scale-Free Error Metrics:
              MeanAbsoluteScaleError(MASE)
    

    Evaluation Metrics for Time-Series:

    Args:
        target: list
        list of true numerical values/number that has to be forecasted 
            --> y_test

        forecast: list
        list of values forecasted corresponding to the target values

        k:
        integer representing the number of features used for prediction in regression

        train:
        list of numerical values/number used for training 
            --> y_train

        bncmrk:
        list of benchmark forecast results for M seasonal period
            --> y_t-M or y_t-1

    """

    def __init__(self, target, forecast, k=1, train=None, bnchmrk=None):
        self.target = target
        self.forecast = forecast
        self.train = train
        self.bnchmrk = bnchmrk
        self.k = k
        self._compute_metrics(target, forecast)

    def _compute_metrics(self, target, forecast, k=1, train=None, bnchmrk=None):
        self.n = len(forecast)
        target, forecast = np.array(target), np.array(forecast)
        self.mean = np.mean(target)

        # Operations used for Scale-Dependent Metrics:
        self.absolute_diff = np.abs(target - forecast)
        self.TSS = np.sum(np.square(forecast - self.mean))
        self.RSS = np.sum(np.square(forecast - target))

        # Operations used for Percentage-Error Metrics
        self.absolute_precentage_diff = np.abs((target - forecast) / target * 100)
        self.absolute_symmetric_diff = np.abs(target - forecast) / (
            (np.abs(target) + np.abs(forecast))
        )

        # Operations used for Relative-Error Metrics
        self.absolute_bnchmrk_diff = np.abs(target - bnchmrk)
        self.absolute_scaled_diff = self.absolute_diff / self.absolute_bnchmrk_diff

        # Operations used for Scale-Free Error Metrics
        self.MAE_in_sample = self.train[:-1] - self.train[1:]

        # Metrics
        # Scale-Dependent Metrics:
        self.MAE = np.mean(self.absolute_diff)
        self.MSE = np.mean(np.square(self.absolute_diff))
        self.RMSE = np.sqrt(self.MSE)
        self.RMSLE = np.log(self.RMSE)
        self.R2 = 1 - self.RSS / self.TSS
        self.adjusted_R2 = 1 - ((self.n - 1) / self.n - k - 1) * (1 - self.R2)

        # Percentage-Error Metrics:
        self.MAPE = np.mean(self.absolute_precentage_diff)
        self.sMAPE = np.mean(self.absolute_symmetric_diff * 100)

        # Relative-Error Metrics:
        self.MdRAE = np.median(self.absolute_scaled_diff)
        self.GMRAE = np.exp(np.mean(np.log(self.absolute_scaled_diff)))

        # Scale-Free Error Metrics:
        self.MASE = self.MAE / self.MAE_in_sample

    def get_MAE(self):
        """
        Returns:
         Mean Absolute Error as float
        """
        return self.MAE

    def get_MSE(self):
        """
        Returns:
         Mean Square Error as float
        """
        return self.MSE

    def get_RSME(self):
        """
        Returns:
         Root Mean Square Error as float
        """
        return self.RSME

    def get_RMSLE(self):
        """
        Returns:
         Root Mean Square Log Error as float
        """
        return self.RMSLE

    def get_R2(self):
        """
        Returns:
         R-Square value as float between optimum (0-1)
        """
        return self.R2

    def get_adjusted_R2(self):
        """
        Returns:
         Adjusted R-Square value as float between (0-1)
        """
        return self.adjusted_R2

    def get_MAPE(self):
        """
        Returns:
         Mean Absolute Percentage Error a percentage as float >0 and optimum <5%
        """
        return self.MAPE

    def get_sMAPE(self):
        """
        Returns:
         Symmetric Mean Absolute Percentage Error a percentage as float in range(0-100)
        """
        return self.sMAPE

    def get_MdRAE(self):
        """
        Returns:
         Median Relative Absolute Error as float, and should be <1
        """
        return self.MdRAE

    def get_GMRAE(self):
        """
        Returns:
         Geoemetric Mean Relative Absolute Error as float, and should be <1
        """
        return self.GMRAE

    def get_MASE(self):
        """
        Returns:
         Mean Absolute Scaled Error as a float, and should be <1
        """
        return self.MASE



from ._object_detection import Metrics
import numpy as np


def batch_to_numpy(predictions, ground_truth):
    """ batch predictions and ground-truth of the return 
    of test process
    
    Paramters:
    ----------
        predictions: dict
            predictions from the return of the test process
            for object detection
        
        ground_truth: dict
            ground-truth from the return of the test process
            for object detection
            
    
    Batch values predictions and ground-truth of detections
    on each image into a numpy array, and gather all 
    fields of predictions together into list
    
    Return:
    -------
        (det_boxes, det_labels, det_scores, true_boxes,  true_labels)
        See parameters for class ObjectDetection for the meaning of
        these parameters
    
    """
    det_boxes, det_labels, det_scores = [], [], []
    true_boxes, true_labels = [], []
    for ix in predictions:
        det_boxes.append(np.array(predictions[ix]["boxes"], dtype=np.float32))
        det_labels.append(np.array(predictions[ix]["objects"], dtype=np.int32))
        det_scores.append(np.array(predictions[ix]["scores"], dtype=np.float32))

        true_boxes.append(np.array(ground_truth[ix]["boxes"], dtype=np.float32))
        true_labels.append(np.array(ground_truth[ix]["objects"], dtype=np.int32))

    return det_boxes, det_labels, det_scores, true_boxes, true_labels



# -*- coding: utf-8 -*-
"""
Created on Fri May 15 23:19:25 2020

@author: arun
"""


from .utils import *
from .imports import *
from .CM import CM as CM


class Metrics(object):
    r"""
    Evaluation Metrics for object detection 

    Args:
        det_boxes: list of numpy.ndarray, 
            one array for each image containing detected objects' bounding boxes.
            Each array is of shape (:math:`M_i`, 4), where :math:`M_i` is the number of objects 
            in the image. Bounding boxes are in xyxy-convention
            All bounding boxes are normalized according to image dimension

        det_labels: list of numpy.ndarray, 
            one array for each image containing detected objects' labels. 
            Each array is of shape (:math:`M_i`)

        det_scores: list of numpy.ndarray, 
            one array for each image 
            containing detected objects' labels' scores.
            Each array is of shape (:math:`M_i`). Object score is 
            the objectness of each detection


        true_boxes: list of numpy.ndarray, 
            one array for each image containing actual objects' bounding boxes.
            Each array is of shape (:math:`N_i`, 4), where :math:`N_i` is the number of 
            ground-truth bbox on the image.
            Bbox in xyxy convention; each bounding box is normalized according 
            to the image dimension

        true_labels: list of numpy.ndarray, 
            one array for each image containing actual objects' labels. 
            Each tensor is of shape :math:`N_i`
        
        num_classes: int
            number of classes (not include background)
                
        threshold: float. Default=0.5
            IoU threshold for a prediction to be considered correct
            with grounth-truth bouding box

    """

    def __init__(
        self,
        det_boxes,
        det_labels,
        det_scores,
        true_boxes,
        true_labels,
        num_classes,
        threshold=0.5,
    ):

        self.det_boxes = det_boxes
        self.det_labels = det_labels
        self.det_scores = det_scores
        self.true_boxes = true_boxes
        self.true_labels = true_labels
        self.num_classes = num_classes
        self.threshold = threshold

        self.reordered_trueboxes = []
        self.reordered_detboxes = []
        self.reordered_detscores = []
        self.confusionMatrix = CM(self.num_classes)
        self.AP = {}
        self.recall = {}
        self.precision = {}
        self._reformatboxes()
        self._evaluate()

    def _reformatboxes(self):
        r"""
            Reformats detection and annotation boxes labelwise 

        """
        for i in range(len(self.true_boxes)):
            self.reordered_trueboxes.append(
                [np.array([]) for _ in range(self.num_classes)]
            )
            self.reordered_detboxes.append(
                [np.array([]) for _ in range(self.num_classes)]
            )
            self.reordered_detscores.append(
                [np.array([]) for _ in range(self.num_classes)]
            )
            currtrueBoxes = self.true_boxes[i]
            currtrueLabels = self.true_labels[i]

            ######## Map Annotation boxes labelwise ####################
            for label in range(self.num_classes):
                self.reordered_trueboxes[-1][label] = currtrueBoxes[
                    currtrueLabels == label
                ]

            if self.det_boxes[i] is not None:
                currdetSort = np.argsort(self.det_scores[i])
                currdetLabels = self.det_labels[i][currdetSort]
                currdetBoxes = self.det_boxes[i][currdetSort]
                currdetScores = self.det_scores[i][currdetSort]

                ######## Map Detection boxes labelwise ####################
                for label in range(self.num_classes):
                    self.reordered_detboxes[-1][label] = currdetBoxes[
                        currdetLabels == label
                    ]
                    self.reordered_detscores[-1][label] = currdetScores[
                        currdetLabels == label
                    ]

    def _evaluate(self):
        r"""
            Computes AP,Precision and recall metrics per class for current model
        
        """
        for label in range(self.num_classes):
            tpBoolmap, detScores = [], []
            nothingness = False
            tp, fp, num_annotations = 0, 0, 0
            for i in range(len(self.reordered_trueboxes)):
                currGT = self.reordered_trueboxes[i][label]
                currDet = self.reordered_detboxes[i][label]
                currDetscores = self.reordered_detscores[i][label]
                detectedGT = []
                num_annotations += currGT.shape[0]

                for box, score in zip(currDet, currDetscores):
                    detScores.append(score)
                    if currGT.shape[0] == 0:
                        tpBoolmap.append(0)
                        fp += 1
                        continue
                    Ious = compute_iou(np.expand_dims(box, axis=0), currGT)
                    assignedGT = np.argmax(
                        Ious, axis=1
                    )  ##### Get the box with maximum overlap
                    currIou = Ious[0, assignedGT]

                    if currIou >= self.threshold and assignedGT not in detectedGT:
                        tp += 1
                        detectedGT.append(assignedGT)
                        tpBoolmap.append(1)
                    else:
                        fp += 1
                        tpBoolmap.append(0)
                ############## Recompute confusion Matrix using current values ###############
                if i == (len(self.reordered_trueboxes) - 1):
                    nothingness = True
                self.recompute_ConfusionMatrix(
                    tp,
                    fp,
                    num_annotations,
                    label,
                    currDet,
                    self.reordered_trueboxes[i],
                    nothingness,
                )

            if num_annotations == 0:
                self.AP[label] = 0
                continue
            tpBoolmap = np.array(tpBoolmap)
            fpBoolmap = np.ones_like(tpBoolmap) - tpBoolmap
            ix = np.argsort(np.array(detScores))
            tpBoolmap = tpBoolmap[ix]
            fpBoolmap = fpBoolmap[ix]
            falsePositives = np.cumsum(fpBoolmap)
            truePositives = np.cumsum(tpBoolmap)
            recall = self.calc_recall(truePositives, num_annotations)
            precision = self.calc_precision(truePositives, falsePositives)

            self.AP[label] = compute_ap(recall, precision)
            self.recall[label] = np.sum(tpBoolmap) / num_annotations
            self.precision[label] = np.sum(tpBoolmap) / (
                np.sum(tpBoolmap) + np.sum(fpBoolmap)
            )

    def getAP(self):
        r"""
        Returns:
            AP per class as a dictionary

        """
        AP = {}
        for c in self.AP:
            AP[c] = self.AP[c]
        return AP

    def getmAP(self):
        r"""
        Computes mAP 
        
        Returns:
             mAP score as float

        """
        return np.mean([ap for c, ap in self.AP.items()]).item()

    def getprecision(self):
        r"""
        Computes Overall Precision 
        
        Returns:
             Precision score as float

        """
        return np.mean(
            [0 if math.isnan(prec) else prec for c, prec in self.precision.items()]
        ).item()

    def getrecall(self):
        r"""
        Computes recall
        
        Returns:
             Recall score as float

        """
        return np.mean([rec for c, rec in self.recall.items()]).item()

    def process_incorrect(self, detections, GTannotations, currlabel):
        r"""
        Function processes detection boxes to count mispredictions and 
        update the confusion matrix
        
        Args:
            detections: Detection boxes of currentlabel 
            GTannotations: Ground Truth annotations of current image 
            currlabel: Label currently being processed
        
        Returns:
            Currently recomputed incorrect detections confusion matrix

        """
        currCM = np.zeros_like(self.confusionMatrix.CM)
        for label in range(self.num_classes):
            if label == currlabel:
                continue
            annotations = GTannotations[label]
            if annotations.shape[0] == 0:
                continue
            detected_annotations = []
            for box in detections:
                Ious = compute_iou(np.expand_dims(box, axis=0), annotations)
                assignedGT = np.argmax(Ious, axis=1)
                maxIou = Ious[0, assignedGT]
                if maxIou >= self.threshold and assignedGT not in detected_annotations:
                    detected_annotations.append(assignedGT)
                    currCM[label, currlabel] += 1
        return currCM

    def calc_precision(self, truePositives, falsePositives):
        r"""
        Fuction computes precision
        
        Args:
            truePositives: True Positives envelope of current label
            falsePositives: False Positives envelope of current label
        
        Returns:
            Precision metrics as float

        """
        return truePositives / np.maximum(
            truePositives + falsePositives, np.finfo(np.float64).eps
        )

    def calc_recall(self, truePositives, nGT):
        r"""
        Fuction computes recall

        Args:
            truePositives: True Positives envelope of current label
            nGT: Number of Ground truth annotations
        
        Returns:
            Recall metrics as float

        """
        return truePositives / nGT

    def recompute_ConfusionMatrix(
        self, tp, fp, num_annotations, label, currDet, currTrue, nothingness=False
    ):

        r"""
        Recomputes confusion matrix imagewise
        
        Args:
            tp: number of True positives
            fp: number of False positives
            num_annotations: number of current ground truth annotations processed
            label: current label being processed
            currDet: current detections being processed
            currTrue: Ground truth boxes for current image
        
        Returns:
            Recall metrics as float

        """
        ############## Update current values into confusion Matrix ###############
        if nothingness:
            fn = num_annotations - tp
            Updatesvals = zip([[label, label], [-1, label], [label, -1]], [tp, fp, fn])
            self.confusionMatrix.updateCM(Updatesvals)
        currCM = self.process_incorrect(currDet, currTrue, label)
        self.confusionMatrix.updateincorrectpredictions(currCM)

    def getCM(self):
        r"""
        Function returns computed Confusion matrix
        
        Returns:
            Computed Confusion Matrix

        """
        return self.confusionMatrix.CM



import os
import json

# ----------------local imports----------------#
from .api_client import APIClient
from .gcp_storage import GCP_Storage
from .console import MainConsole

# ----------------frame works------------------#
import pandas as pd


class AnnotationStreaming:
    '''
    A class that returns the labels and annotations from the cloud bucket and creates a dataset.
    Returns datapoints along with the labels as numpy array.
    '''
    def __init__(self, token, root, indices) -> None:
        self.token = token
        self.root = root
        self.indices = indices
        self.console = MainConsole()
        dir_path = os.path.dirname(os.path.realpath(__file__))
        
        with open(os.path.join(dir_path, "config.json"), "r") as f:
            self.config = json.load(f)
        
        self.api_client = APIClient(
            backend_url=self.config["backend_ip"], token=self.token)
        
        self.gcp_client = GCP_Storage(exp_token=token)


    def __pull_annotations(self):
        self.console.info("PULLING ANNOTATION FROM CLOUD")
        
        error, data = self.api_client.POST_REQUEST(
            end_point=f"/v2/get_annotation_data",
            payload={
                    "token": self.token,
                    "selected_indices": self.indices
                }
        )
        
        if error:
            self.console.error(
                "SOMETHING WENT WRONG WITH ALECTIO ANNOTATIONS LOADER.\nCONTACT ALECTIO ADMIN"
            )

        json_data = json.loads(data["data"]["dataframe"])
        dataframe = pd.DataFrame(json_data)

        if data["data"]["task_type"].lower() in ["object_detection", "image_classification"]:
            self.datapoints = dataframe['file_path'].to_list()
            self.annotations = dataframe['annotations'].to_list()
        elif data["data"]["task_type"].lower() == "text_classification":
            self.datapoints = dataframe['text'].to_list()
            self.annotations = dataframe['annotations'].to_list()
        else:
            raise Exception(f"{data['data']['task_type']} TASK TYPE IS NOT SUPPORTED YET")


    def get_dataset(self):
        self.__pull_annotations()
        return self.datapoints, self.annotations


import setuptools


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


install_reqs = parse_requirements("./requirements.txt")

with open("README.md", "r") as fh:
    long_description = fh.read()

reqs = install_reqs
setuptools.setup(
    name="alectio_sdk",
    version="1.0.31",
    author="Alectio",
    author_email="admin@alectio.com",
    url='https://github.com/alectio/SDK',
    description="Integrate customer side ML application with the Alectio Platform",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    entry_points = {
        'console_scripts': ['alectio=alectio_sdk.cli.cmd:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Logging',
        'Topic :: System :: Monitoring'
    ],
    python_requires=">=3.6",
    install_requires=install_reqs,
    package_data={"": ["config.json"]},
    include_package_data=True,
)



import tensorflow as tf
from tensorflow_probability import distributions as tfd
import numpy as np

def bayesian_inference(predictions):
    data = predictions
    # Define the prior distribution
    prior_mean = tfd.Normal(loc=0, scale=1)
    prior_stddev = tfd.HalfNormal(scale=1)

    # Define the likelihood function
    def likelihood(mu, sigma):
        dist = tfd.Normal(loc=mu, scale=sigma)
        return tf.reduce_prod(dist.prob(data))

    # Define the posterior distribution
    @tf.function
    def posterior(mu, sigma):
        return likelihood(mu, sigma) * prior_mean.prob(mu) * prior_stddev.prob(sigma)

    # Define the optimizer
    optimizer = tf.optimizers.Adam()
    # Optimize the posterior distribution
    mu = tf.Variable(0.0)
    sigma = tf.Variable(1.0)

    for i in range(1000):
        with tf.GradientTape() as tape:
            loss = -tf.math.log(posterior(mu, sigma))
        gradients = tape.gradient(loss, [mu, sigma])
        optimizer.apply_gradients(zip(gradients, [mu, sigma]))

    # Get the posterior distribution
    posterior_mean = mu.numpy()
    posterior_stddev = sigma.numpy()
    return posterior_mean, posterior_stddev


import typer


class MainConsole:
    def __init__(self) -> None:
        pass

    def error(self, msg: str):
        msg = "❌  " + msg + "\n"
        message = typer.style(msg, fg=typer.colors.RED)
        typer.echo(message)

    def success(self, msg: str):
        msg = "✅  " + msg + "\n"
        message = typer.style(msg, fg=typer.colors.GREEN)
        typer.echo(message)

    def info(self, msg: str):
        msg = "ℹ️  " + msg + "\n"
        message = typer.style(msg, fg=typer.colors.MAGENTA)
        typer.echo(message)

    def warning(self, msg: str):
        msg = "⚠️  " + msg + "\n"
        message = typer.style(msg, fg=typer.colors.YELLOW)
        typer.echo(message)

    def message(self, msg: str):
        message = typer.style(msg, fg=typer.colors.CYAN)
        typer.echo(message)



# -*- coding: utf-8 -*-
"""
@author: arun
"""
import logging
import numpy as np
from .SegCM import SegCM as CM
from collections import OrderedDict
from .rangeutils import \
    SemLaserScan  #####  Requirement for rangenet implementation Source: http://www.ipb.uni-bonn.de/wp-content/papercite-data/pdf/milioto2019iros.pdf


def fast_hist(label_true, label_pred, n_class):
    mask = (label_true >= 0) & (label_true < n_class)
    hist = np.bincount(
        n_class * label_true[mask].astype(int) + label_pred[mask].astype(int), minlength=n_class ** 2
    ).reshape(n_class, n_class)
    return hist


def segmentation_metrics(ground_truth, predictions, n_classes):
    """
    ground_truth: list of 2d labels
    predictions: list of 2d predicted labels
    """
    confusion_matrix = np.zeros((n_classes, n_classes))

    for pred, gt in zip(predictions, ground_truth):
        confusion_matrix += fast_hist(gt.flatten(), pred.flatten(), n_classes)

    FP = confusion_matrix.sum(axis=0) - np.diag(confusion_matrix)
    FN = confusion_matrix.sum(axis=1) - np.diag(confusion_matrix)
    TP = confusion_matrix.diagonal()
    TN = confusion_matrix.sum() - (FP + FN + TP)

    precision = np.divide(TP, (TP + FP), out=np.zeros_like(TP), where=(TP + FP) != 0)
    recall = np.divide(TP, (TP + FN), out=np.zeros_like(TP), where=(TP + FN) != 0)

    f1_score = 2 * np.divide((precision * recall), (precision + recall), out=np.zeros_like(precision * recall),
                             where=(precision + recall) != 0)

    hist = confusion_matrix
    acc = np.diag(hist).sum() / hist.sum()
    acc_cls = np.diag(hist) / hist.sum(axis=1)
    iu = np.diag(hist) / (hist.sum(axis=1) + hist.sum(axis=0) - np.diag(hist))
    mean_iu = np.nanmean(iu)
    freq = hist.sum(axis=1) / hist.sum()
    # TODO: check with prateek on if freqwacc is okay
    fwavacc = (freq[freq > 0] * iu[freq > 0]).sum()
    classwise_iou = dict(zip(range(n_classes), iu))
    classwise_pixel_acc = {k: v for (k, v) in enumerate(acc_cls.tolist())}
    classwise_dice = {k: v for (k, v) in enumerate(f1_score.tolist())}

    metrics = {
        "confusion_matrix": confusion_matrix.tolist(),
        "pixel_acc": acc,
        "freqw_iou": fwavacc,
        "mean_iou": mean_iu,
        "dice": f1_score.mean(),
        "classwise_pixel_acc": classwise_pixel_acc,
        "classwise_iou": classwise_iou,
        "classwise_dice": classwise_dice
    }

    return metrics


class SegMetrics(object):
    """
  Segmentation Evaluation
  Current Support :
  2D
  3D - Support only Rangenet as seen on http://www.ipb.uni-bonn.de/wp-content/papercite-data/pdf/milioto2019iros.pdf

  """

    def __init__(self,
                 n_classes,
                 labels,
                 return_2D=True,
                 return_3D=True,
                 rangenet=True,
                 default_ranges=None,
                 include_ranges=True,
                 range_labels=OrderedDict()
                 ):

        self.n_classes = n_classes
        self.labels = labels

        if rangenet:
            logging.info(
                "Initializing Segmentation evaluation for {} classes with range projection from Rangenet ".format(
                    n_classes))
            if include_ranges is True and default_ranges is None:
                raise TypeError("Default ranges cannot be None when include_ranges is set to True")
            self.default_ranges = default_ranges
            self.include_ranges = include_ranges
        else:
            logging.info("Initializing Segmentation evaluation for {} classes ".format(n_classes))

    def num_classes(self):
        return self.n_classes

    def compute(self):
        self.max_range_conf.add_single_comparison(gt_scan.sem_label,
                                                  pred_scan.sem_label)

    def _reset3D(self):
        logging.info("Resetting metrics evaluation for segmentation")
        self.pclCM = CM(nClasses=nclasses, labels=self.labels, task="Segmentation")  # 3D
        if include_ranges:
            self.pclrangeCM = OrderedDict(
                (range_spec, CM(nClasses=nclasses, labels=self.labels, task="Segmentation")) for range_spec in
                self.default_ranges)
            self.rangemiou = OrderedDict((range_spec, {}) for range_spec in self.default_ranges)
            self.rangeiou = OrderedDict((range_spec, {}) for range_spec in self.default_ranges)
            self.rangedice = OrderedDict((range_spec, {}) for range_spec in self.default_ranges)
            self.rangemdice = OrderedDict((range_spec, {}) for range_spec in self.default_ranges)
            self.rangeacc = OrderedDict((range_spec, {}) for range_spec in self.default_ranges)
            self.rangemacc = OrderedDict((range_spec, {}) for range_spec in self.default_ranges)
        else:
            self.pclrangeCM = {}

    def _reset2D(self):
        logging.info("Resetting metrics evaluation for segmentation")
        self.imageCM = CM(nClasses=nclasses, labels=self.labels, task="Segmentation")  # 2D

    def evaluate3D(self, gt, preds, rangelabels=OrderedDict()):
        """
    3D segmentation evaluyations only includes support for Rangenet types

    """
        logging.info("Resetting metrics evaluation for 3D rangenet based segmentation")
        self._reset3D()

        # Evaluating range wise and non rangewise confusion matrix
        for k in gt.keys():
            for groundtruth, predicted in zip(gt[k], pred[k]):
                self.pclCM.updateCM(groundtruth, predicted)
            if self.include_ranges:
                if not rangelabels:
                    raise ValueError(
                        "Range labels cannot be empty for range based evaluations , OrderedDict in format [GTlabels, predlabels] is required")
                for rangeval, rangeCM in self.pclrangeCM.items():
                    rangeGT, rangepreds = rangelabels[k][rangeval]
                    self.pclrangeCM[k][rangeval].updateCM(rangeGT, rangepreds)

    def evaluate2D(self, gt, preds):
        """
    3D segmentation evaluyations only includes support for Rangenet types

    """
        logging.info("Resetting metrics evaluation for 2D Segmentation")
        self._reset2D()
        for k in gt.keys():
            for groundtruth, predicted in zip(gt[k], pred[k]):
                self.imageCM.updateCM(groundtruth, predicted)

    def get3DrangeCM(self):
        if self.include_ranges:
            return self.pclrangeCM
        else:
            return None

    def get3DCM(self):
        return self.pclCM

    def get3Drangeacc(self):
        for range_spec in self.self.rangeacc.keys():
            self.self.rangeacc[range_spec] = self.calcaccuracy(self, self.pclrangeCM[range_spec])
        return self.rangeacc

    def get3Dacc(self):
        return self.calcaccuracy(self, self.pclCM)

    def get3Dmacc(self):
        return self.calcmaccuracy(self, self.pclCM)

    def get3Drangemacc(self):
        for range_spec in self.rangemacc.keys():
            self.rangemacc[range_spec] = self.calcmaccuracy(self, self.pclrangeCM[range_spec])
        return self.rangemacc

    def get3DrangeIOU(self):
        for range_spec in self.rangeiou.keys():
            self.rangeiou[range_spec] = self.calcIOU(self, self.pclrangeCM[range_spec])
        return self.rangeiou

    def get3DIOU(self):
        return self.calcIOU(self, self.pclCM)

    def get3DmIOU(self):
        return self.calcMIOU(self, self.pclCM)

    def get3DrangemIOU(self):
        for range_spec in self.rangemiou.keys():
            self.rangemiou[range_spec] = self.calcMIOU(self, self.pclrangeCM[range_spec])
        return self.rangemiou

    def get3DDICE(self):
        return self.calcDICE(self, self.pclCM)

    def get3DrangeDICE(self):
        for range_spec in self.rangedice.keys():
            self.rangedice[range_spec] = self.calcDICE(self, self.pclrangeCM[range_spec])
        return self.rangedice

    def get3DmDICE(self):
        return self.calcmDICE(self, self.pclCM)

    def get3DrangemDICE(self):
        for range_spec in self.rangemdice.keys():
            self.rangemdice[range_spec] = self.calcmDICE(self, self.pclrangeCM[range_spec])
        return self.rangemdice

    def get2DCM(self):
        return self.imageCM

    def get2DAcc(self):
        return self.calcaccuracy(self, self.imageCM)

    def get2DmAcc(self):
        return self.calcmaccuracy(self, self.imageCM)

    def get2DIOU(self):
        return self.calcIOU(self, self.imageCM)

    def get2DmIOU(self):
        return self.calcMIOU(self, self.imageCM)

    def get2DDICE(self):
        return self.calcDICE(self, self.imageCM)

    def get2DmDICE(self):
        return self.calcmDICE(self, self.imageCM)

    def get2DfwIOU(self):
        return self.calcfwIOU(self, self.imageCM)

    def calcfwIOU(self, inputCM):
        intersection = np.diagonal(inputCM)
        # Calculate area of Union , Union =
        union = (np.sum(inputCM, axis=0) + np.sum(inputCM, axis=1)) - intersection
        classwiseiou = (intersection / union)
        return np.divide(np.sum(np.multiply(np.sum(inputCM, axis=0), classwiseiou)), np.sum(inputCM))

    def calcIOU(self, inputCM):
        """
    Calculates and returns classwise IOU for the input segmentation labels and predictions


    IOU  = (Area of Intersection)/(Area of union)
    """

        # Calculate area of Intersection
        intersection = np.diagonal(inputCM)
        # Calculate area of Union , Union =
        union = (np.sum(inputCM, axis=0) + np.sum(inputCM, axis=1)) - intersection
        classwiseiou = (intersection / union)
        classwiseioudict = {}

        for c in range(self.n_classes):
            classwiseioudict[c] = classwiseiou[c]

        return classwiseioudict

    def calcMIOU(self, inputCM):
        """
    Calculates and returns IOU for the input segmentation labels and predictions
    """
        # Calculate area of Intersection

        intersection = np.diagonal(inputCM)
        # Calculate area of Union ,
        union = (np.sum(inputCM, axis=0) + np.sum(inputCM, axis=1)) - intersection

        return np.nanmean(intersection / union)

    def calcDICE(self, inputCM):
        # Calculate area of Intersection
        intersection = np.diagonal(inputCM)

        # Calculate total pixels combined
        totalpixels = (np.sum(inputCM, axis=0) + np.sum(inputCM, axis=1))

        classwiseDICE = ((2 * intersection) / totalpixels)
        classwiseDicedict = {}

        for c in range(self.n_classes):
            classwiseDicedict[c] = classwiseDICE[c]

        return classwiseDicedict

    def calcmDICE(self, inputCM):
        # Calculate area of Intersection
        intersection = np.diagonal(inputCM)

        # Calculate total pixels combined
        totalpixels = (np.sum(inputCM, axis=0) + np.sum(inputCM, axis=1))

        return np.nanmean((2 * intersection) / totalpixels)

    def calcaccuracy(self, inputCM):
        # Calculate area of Intersection
        intersection = np.diagonal(inputCM)

        # Calculate total pixels combined
        totalGTpixels = (np.sum(inputCM, axis=1))

        classwiseaccuracy = (intersection / totalGTpixels)
        classwiseaccdict = {}

        for c in range(self.n_classes):
            classwiseaccdict[c] = classwiseaccuracy[c]

        return classwiseaccdict

    def calcmaccuracy(self, inputCM):
        # Calculate area of Intersection
        intersection = np.diagonal(inputCM)

        # Calculate total pixels combined
        totalGTpixels = (np.sum(inputCM, axis=1))

        return np.nanmean((intersection / totalGTpixels))



import math

from sklearn.metrics import hamming_loss
from sklearn.metrics import accuracy_score
from sklearn.metrics import multilabel_confusion_matrix

class MultiLabelClassificationMetrics(object):
    r"""
    Evaluation Metrics for multilabel classification
    Args:
        predictions: list,
            list of classes predicted for each record

        ground_truth: list,
            list of ground_truth classes for each record
    """

    def __init__(self, predictions, ground_truth):
        self.predictions = predictions
        self.ground_truth = ground_truth
        self._compute_metrics(predictions, ground_truth)

    def _compute_metrics(self, predictions, ground_truth):
        self.accuracy_result = accuracy_score(ground_truth, predictions)
        self.hamming_l= hamming_loss(ground_truth, predictions)

        multilabel_confusion_mat = multilabel_confusion_matrix(ground_truth, predictions)

        self.output = {}
        for k, row in enumerate(multilabel_confusion_mat):
            tn, fp, fn, tp = row.ravel()
            self.output[k] = {'TN': tn , 'FP': fp, 'FN': fn, 'TP': tp}

        TP, FP, TN, FN = 0, 0, 0, 0 # micro
        self.precision_macro, self.recall_macro = 0.0, 0.0 # macro
        for c, mat in self.output.items():
            TP += mat['TP']
            FP += mat['FP']
            TN += mat['TN']
            FN += mat['FN']

            p = mat['TP'] / (mat['TP'] + mat['FP'])
            r = mat['TP'] / (mat['TP'] + mat['FN'])

            self.precision_macro = self.precision_macro + p if not math.isnan(p) else self.precision_macro
            self.recall_macro = self.recall_macro + r if not math.isnan(r) else self.recall_macro
        
        self.precision_micro = TP / (TP + FP)
        self.recall_micro = TP / (TP + FN)
        self.f1_micro = 2 * (self.precision_micro * self.recall_micro) / (self.precision_micro + self.recall_micro)

        self.precision_micro = self.precision_micro if not math.isnan(self.precision_micro) else 0
        self.recall_micro = self.recall_micro if not math.isnan(self.recall_micro) else 0
        self.f1_micro = self.f1_micro if not math.isnan(self.f1_micro) else 0

        self.precision_macro = self.precision_macro / len(multilabel_confusion_mat)
        self.recall_macro = self.recall_macro / len(multilabel_confusion_mat)

        self.f1_macro = 2 * (self.precision_macro * self.recall_macro) / (self.precision_macro + self.recall_macro)
        self.f1_macro = self.f1_macro if not math.isnan(self.f1_macro) else 0

    def get_accuracy(self):
        r"""
        Returns:
            accuracy as a float
        """
        return self.accuracy_result

    def get_f1_score_micro(self):
        r"""
        Returns:
            F1 score micro
        """
        return self.f1_micro

    def get_f1_score_macro(self):
        r"""
        Returns:
            F1 score macro
        """
        return self.f1_macro

    def get_precision_micro(self):
        r"""
        Returns:
            precision micro
        """
        return self.precision_micro

    def get_precision_macro(self):
        r"""
        Returns:
            precision macro
        """
        return self.precision_macro

    def get_recall_micro(self):
        r"""
        Returns:
            recall micro
        """
        return self.recall_micro

    def get_recall_macro(self):
        r"""
        Returns:
            recall macro
        """
        return self.recall_macro

    def get_confusion_matrix(self):
        r"""
        Returns:
            confusion matrix as a list
        """
        return self.output

    def get_hamming_loss(self):
        r"""
        Returns:
            hamming loss
        """
        return self.hamming_l



from typing import Any, Dict
from google.cloud import storage
from google.oauth2 import service_account
import json
import pickle
from PIL import Image
import numpy as np
import os
import io
import shutil
from rich.console import Console
from alectio_sdk.sdk.utils import get_storage_credentials


console = Console()


   
class GCP_Storage:
    exp_token : str
    gcpStorageClient: None

    # def init :- experimenttoken
    def __init__(self, exp_token):
        self.exp_token = exp_token
        storage_info=get_storage_credentials(exp_token=exp_token)
    
        storage_info = json.loads(storage_info)
        credentials = service_account.Credentials.from_service_account_info(storage_info)
        self.gcpStorageClient = storage.Client(project=None, credentials=credentials)
        

    def download_file(
        self, storage_bucket: str, object_key: str, format: str, dest_path: str
    ):
        bucket = self.gcpStorageClient.get_bucket(storage_bucket)
        blob = bucket.blob(object_key)
        if format in ["jpg", "png", "jpeg"]:
            os.makedirs(os.path.dirname(dest_path), exist_ok=True)
            blob.download_to_filename(dest_path)
        if format == "zip":
            done = False
            with console.status("[bold green]Loading dataset from cloud...") as status:
                while not done:
                    blob.download_to_filename(dest_path + ".zip")
                    shutil.unpack_archive(dest_path + ".zip", dest_path)
                    done = True

    def read_file(self, storage_bucket: str, object_key: str, format: str):
        # print('we are hitting this function!!!!')
        # print(object_key)
        # print(storage_bucket)
        bucket = self.gcpStorageClient.get_bucket(storage_bucket)
        blob = bucket.blob(object_key)

        if format == "json":
            data = json.loads(blob.download_as_string(client=None))
        if format == "pkl":
            data = pickle.loads(blob.download_as_string(client=None))
        if format == "text":
            data = blob.download_as_string(client=None)
        return data

    def upload_object(
        self, storage_bucket: str, object: Any, object_key: str, format: str
    ):
        bucket = self.gcpStorageClient.get_bucket(storage_bucket)
        blob = bucket.blob(object_key)

        if format == "json":
            blob.upload_from_string(data=json.dumps(object))
        if format == "pkl":
            blob.upload_from_string(data=pickle.dumps(object))
        if format == "txt":
            blob.upload_from_string(data=object)
        if format in ["png", "jpg"]:
            img = Image.fromarray(np.uint8(object)).convert("RGB")
            blob.upload_from_string(img.tobytes(), content_type="image/jpeg")

    def upload_file(self, storage_bucket: str, file_path: str, object_key: str):
        bucket = self.gcpStorageClient.get_bucket(storage_bucket)
        blob = bucket.blob(object_key)
        blob.upload_from_filename(file_path)
